#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools
import inspect

from oslo_log import log as logging
from oslo_messaging import rpc

LOG = logging.getLogger(__name__)


def asynchronous(function):
    """Decorator for MessageProcessor methods to make them asynchronous.

    To use, simply call the method as usual. Instead of being executed
    immediately, it will be placed on the queue for the MessageProcessor and
    run on a future iteration of the event loop.
    """

    sig = inspect.signature(function)
    arg_names = [name for name, param in sig.parameters.items()
                 if param.kind == param.POSITIONAL_OR_KEYWORD]
    MessageData = collections.namedtuple(function.__name__, arg_names[1:])

    @functools.wraps(function)
    def call_or_send(processor, *args, **kwargs):
        if len(args) == 1 and not kwargs and isinstance(args[0], MessageData):
            try:
                return function(processor, **args[0]._asdict())
            except rpc.dispatcher.ExpectedException as exc:
                LOG.error('[%s] Exception in "%s": %s',
                          processor.name, function.__name__, exc.exc_info[1],
                          exc_info=exc.exc_info)
                raise
            except Exception as exc:
                LOG.exception('[%s] Exception in "%s": %s',
                              processor.name, function.__name__, exc)
                raise
        else:
            data = inspect.getcallargs(function, processor, *args, **kwargs)
            data.pop(arg_names[0])  # lose self
            return processor.queue.send(function.__name__,
                                        MessageData(**data))

    call_or_send.MessageData = MessageData
    return call_or_send


class MessageProcessor(object):

    queue = None

    def __init__(self, name):
        self.name = name

    def __call__(self):
        message = self.queue.get()
        if message is None:
            LOG.debug('[%s] No messages', self.name)
            return False

        try:
            method = getattr(self, message.name)
        except AttributeError:
            LOG.error('[%s] Bad message name "%s"' % (self.name,
                                                      message.name))
            raise
        else:
            LOG.info('[%s] %r' % (self.name, message.data))

        method(message.data)
        return True

    @asynchronous
    def noop(self, count=1):
        """Insert <count> No-op operations in the message queue."""
        assert isinstance(count, int)
        if count > 1:
            self.queue.send_priority('noop',
                                     self.noop.MessageData(count - 1))

    @asynchronous
    def _execute(self, func):
        """Insert a function call in the message queue.

        The function takes no arguments, so use functools.partial to curry the
        arguments before passing it here.
        """
        func()

    def call(self, func, *args, **kwargs):
        """Insert a function call in the message queue."""
        self._execute(functools.partial(func, *args, **kwargs))

    def clear(self):
        """Delete all the messages from the queue."""
        self.queue.clear()


__all__ = ['MessageProcessor', 'asynchronous']



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import uuid


from heat.common import exception
from heat.common.i18n import _
from heat.engine.cfn import functions
from heat.engine import environment
from heat.engine import function
from heat.engine import resource
from heat.engine import rsrc_defn
from heat.engine import stack
from heat.engine import stk_defn
from heat.engine import template
from heat.tests import common
from heat.tests import utils


class TestFunction(function.Function):
    def validate(self):
        if len(self.args) < 2:
            raise TypeError(_('Need more arguments'))

    def dependencies(self, path):
        return ['foo', 'bar']

    def result(self):
        return 'wibble'


class NullFunction(function.Function):
    def result(self):
        return Ellipsis


class TestFunctionKeyError(function.Function):
    def result(self):
        raise TypeError


class TestFunctionValueError(function.Function):
    def result(self):
        raise ValueError


class TestFunctionResult(function.Function):
    def result(self):
        return super(TestFunctionResult, self).result()


class FunctionTest(common.HeatTestCase):
    def test_equal(self):
        func = TestFunction(None, 'foo', ['bar', 'baz'])
        self.assertTrue(func == 'wibble')
        self.assertTrue('wibble' == func)

    def test_not_equal(self):
        func = TestFunction(None, 'foo', ['bar', 'baz'])
        self.assertTrue(func != 'foo')
        self.assertTrue('foo' != func)

    def test_equal_func(self):
        func1 = TestFunction(None, 'foo', ['bar', 'baz'])
        func2 = TestFunction(None, 'blarg', ['wibble', 'quux'])
        self.assertTrue(func1 == func2)

    def test_function_str_value(self):
        func1 = TestFunction(None, 'foo', ['bar', 'baz'])
        expected = '%s %s' % ("<heat.tests.test_function.TestFunction",
                              "{foo: ['bar', 'baz']} -> 'wibble'>")
        self.assertEqual(expected, str(func1))

    def test_function_stack_reference_none(self):
        func1 = TestFunction(None, 'foo', ['bar', 'baz'])
        self.assertIsNone(func1.stack)

    def test_function_exception_key_error(self):
        func1 = TestFunctionKeyError(None, 'foo', ['bar', 'baz'])
        expected = '%s %s' % ("<heat.tests.test_function.TestFunctionKeyError",
                              "{foo: ['bar', 'baz']} -> ???>")
        self.assertEqual(expected, str(func1))

    def test_function_eq_exception_key_error(self):
        func1 = TestFunctionKeyError(None, 'foo', ['bar', 'baz'])
        func2 = TestFunctionKeyError(None, 'foo', ['bar', 'baz'])
        result = func1.__eq__(func2)
        self.assertEqual(result, NotImplemented)

    def test_function_ne_exception_key_error(self):
        func1 = TestFunctionKeyError(None, 'foo', ['bar', 'baz'])
        func2 = TestFunctionKeyError(None, 'foo', ['bar', 'baz'])
        result = func1.__ne__(func2)
        self.assertEqual(result, NotImplemented)

    def test_function_exception_value_error(self):
        func1 = TestFunctionValueError(None, 'foo', ['bar', 'baz'])
        expected = '%s %s' % (
            "<heat.tests.test_function.TestFunctionValueError",
            "{foo: ['bar', 'baz']} -> ???>")
        self.assertEqual(expected, str(func1))

    def test_function_eq_exception_value_error(self):
        func1 = TestFunctionValueError(None, 'foo', ['bar', 'baz'])
        func2 = TestFunctionValueError(None, 'foo', ['bar', 'baz'])
        result = func1.__eq__(func2)
        self.assertEqual(result, NotImplemented)

    def test_function_ne_exception_value_error(self):
        func1 = TestFunctionValueError(None, 'foo', ['bar', 'baz'])
        func2 = TestFunctionValueError(None, 'foo', ['bar', 'baz'])
        result = func1.__ne__(func2)
        self.assertEqual(result, NotImplemented)

    def test_function_abstract_result(self):
        func1 = TestFunctionResult(None, 'foo', ['bar', 'baz'])
        expected = '%s %s -> %s' % (
            "<heat.tests.test_function.TestFunctionResult",
            "{foo: ['bar', 'baz']}",
            "{'foo': ['bar', 'baz']}>")
        self.assertEqual(expected, str(func1))

    def test_copy(self):
        func = TestFunction(None, 'foo', ['bar', 'baz'])
        self.assertEqual({'foo': ['bar', 'baz']}, copy.deepcopy(func))


class ResolveTest(common.HeatTestCase):
    def test_resolve_func(self):
        func = TestFunction(None, 'foo', ['bar', 'baz'])

        result = function.resolve(func)

        self.assertEqual('wibble', result)
        self.assertIsInstance(result, str)

    def test_resolve_dict(self):
        func = TestFunction(None, 'foo', ['bar', 'baz'])
        snippet = {'foo': 'bar', 'blarg': func}

        result = function.resolve(snippet)

        self.assertEqual({'foo': 'bar', 'blarg': 'wibble'}, result)
        self.assertIsNot(result, snippet)

    def test_resolve_list(self):
        func = TestFunction(None, 'foo', ['bar', 'baz'])
        snippet = ['foo', 'bar', 'baz', 'blarg', func]

        result = function.resolve(snippet)

        self.assertEqual(['foo', 'bar', 'baz', 'blarg', 'wibble'], result)
        self.assertIsNot(result, snippet)

    def test_resolve_all(self):
        func = TestFunction(None, 'foo', ['bar', 'baz'])
        snippet = ['foo', {'bar': ['baz', {'blarg': func}]}]

        result = function.resolve(snippet)

        self.assertEqual(['foo', {'bar': ['baz', {'blarg': 'wibble'}]}],
                         result)
        self.assertIsNot(result, snippet)

    def test_resolve_func_with_null(self):
        func = NullFunction(None, 'foo', ['bar', 'baz'])

        self.assertIsNone(function.resolve(func))
        self.assertIs(Ellipsis, function.resolve(func, nullable=True))

    def test_resolve_dict_with_null(self):
        func = NullFunction(None, 'foo', ['bar', 'baz'])
        snippet = {'foo': 'bar', 'baz': func, 'blarg': 'wibble'}

        result = function.resolve(snippet)

        self.assertEqual({'foo': 'bar', 'blarg': 'wibble'}, result)

    def test_resolve_list_with_null(self):
        func = NullFunction(None, 'foo', ['bar', 'baz'])
        snippet = ['foo', func, 'bar']

        result = function.resolve(snippet)

        self.assertEqual(['foo', 'bar'], result)


class ValidateTest(common.HeatTestCase):
    def setUp(self):
        super(ValidateTest, self).setUp()
        self.func = TestFunction(None, 'foo', ['bar', 'baz'])

    def test_validate_func(self):
        self.assertIsNone(function.validate(self.func))
        self.func = TestFunction(None, 'foo', ['bar'])
        self.assertRaisesRegex(exception.StackValidationFailed,
                               'foo: Need more arguments',
                               function.validate, self.func)

    def test_validate_dict(self):
        snippet = {'foo': 'bar', 'blarg': self.func}
        function.validate(snippet)

        self.func = TestFunction(None, 'foo', ['bar'])
        snippet = {'foo': 'bar', 'blarg': self.func}
        self.assertRaisesRegex(exception.StackValidationFailed,
                               'blarg.foo: Need more arguments',
                               function.validate, snippet)

    def test_validate_list(self):
        snippet = ['foo', 'bar', 'baz', 'blarg', self.func]
        function.validate(snippet)

        self.func = TestFunction(None, 'foo', ['bar'])
        snippet = {'foo': 'bar', 'blarg': self.func}
        self.assertRaisesRegex(exception.StackValidationFailed,
                               'blarg.foo: Need more arguments',
                               function.validate, snippet)

    def test_validate_all(self):
        snippet = ['foo', {'bar': ['baz', {'blarg': self.func}]}]
        function.validate(snippet)

        self.func = TestFunction(None, 'foo', ['bar'])
        snippet = {'foo': 'bar', 'blarg': self.func}
        self.assertRaisesRegex(exception.StackValidationFailed,
                               'blarg.foo: Need more arguments',
                               function.validate, snippet)


class DependenciesTest(common.HeatTestCase):
    func = TestFunction(None, 'test', None)

    scenarios = [
        ('function', dict(snippet=func)),
        ('nested_map', dict(snippet={'wibble': func})),
        ('nested_list', dict(snippet=['wibble', func])),
        ('deep_nested', dict(snippet=[{'wibble': ['wibble', func]}])),
    ]

    def test_dependencies(self):
        deps = list(function.dependencies(self.snippet))
        self.assertIn('foo', deps)
        self.assertIn('bar', deps)
        self.assertEqual(2, len(deps))


class ValidateGetAttTest(common.HeatTestCase):
    def setUp(self):
        super(ValidateGetAttTest, self).setUp()

        env = environment.Environment()
        env.load({u'resource_registry':
                  {u'OS::Test::GenericResource': u'GenericResourceType'}})

        env.load({u'resource_registry':
                  {u'OS::Test::FakeResource': u'OverwrittenFnGetAttType'}})

        tmpl = template.Template({"HeatTemplateFormatVersion": "2012-12-12",
                                  "Resources": {
                                      "test_rsrc": {
                                          "Type": "OS::Test::GenericResource"
                                      },
                                      "get_att_rsrc": {
                                          "Type": "OS::Heat::Value",
                                          "Properties": {
                                              "value": {
                                                  "Fn::GetAtt": ["test_rsrc",
                                                                 "Foo"]
                                              }
                                          }
                                      }
                                  }},
                                 env=env)
        self.stack = stack.Stack(
            utils.dummy_context(), 'test_stack',
            tmpl,
            stack_id=str(uuid.uuid4()))
        self.rsrc = self.stack['test_rsrc']
        self.stack.validate()

    def test_resource_is_appear_in_stack(self):
        func = functions.GetAtt(self.stack.defn, 'Fn::GetAtt',
                                [self.rsrc.name, 'Foo'])
        self.assertIsNone(func.validate())

    def test_resource_is_not_appear_in_stack(self):
        self.stack.remove_resource(self.rsrc.name)

        func = functions.GetAtt(self.stack.defn, 'Fn::GetAtt',
                                [self.rsrc.name, 'Foo'])
        ex = self.assertRaises(exception.InvalidTemplateReference,
                               func.validate)
        self.assertEqual('The specified reference "test_rsrc" (in unknown) '
                         'is incorrect.', str(ex))

    def test_resource_no_attribute_with_default_fn_get_att(self):
        res_defn = rsrc_defn.ResourceDefinition('test_rsrc',
                                                'ResWithStringPropAndAttr')
        self.rsrc = resource.Resource('test_rsrc', res_defn, self.stack)
        self.stack.add_resource(self.rsrc)
        stk_defn.update_resource_data(self.stack.defn, self.rsrc.name,
                                      self.rsrc.node_data())
        self.stack.validate()

        func = functions.GetAtt(self.stack.defn, 'Fn::GetAtt',
                                [self.rsrc.name, 'Bar'])
        ex = self.assertRaises(exception.InvalidTemplateAttribute,
                               func.validate)
        self.assertEqual('The Referenced Attribute (test_rsrc Bar) '
                         'is incorrect.', str(ex))

    def test_resource_no_attribute_with_overwritten_fn_get_att(self):
        res_defn = rsrc_defn.ResourceDefinition('test_rsrc',
                                                'OS::Test::FakeResource')
        self.rsrc = resource.Resource('test_rsrc', res_defn, self.stack)
        self.rsrc.attributes_schema = {}
        self.stack.add_resource(self.rsrc)
        stk_defn.update_resource_data(self.stack.defn, self.rsrc.name,
                                      self.rsrc.node_data())
        self.stack.validate()

        func = functions.GetAtt(self.stack.defn, 'Fn::GetAtt',
                                [self.rsrc.name, 'Foo'])
        self.assertIsNone(func.validate())

    def test_get_attr_without_attribute_name(self):
        ex = self.assertRaises(ValueError, functions.GetAtt,
                               self.stack.defn, 'Fn::GetAtt', [self.rsrc.name])
        self.assertEqual('Arguments to "Fn::GetAtt" must be '
                         'of the form [resource_name, attribute]',
                         str(ex))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine import rsrc_defn
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

qos_policy_template = '''
heat_template_version: 2016-04-08
description: This template to define a neutron qos policy.
resources:
  my_qos_policy:
    type: OS::Neutron::QoSPolicy
    properties:
      description: a policy for test
      shared: true
      tenant_id: d66c74c01d6c41b9846088c1ad9634d0
'''

bandwidth_limit_rule_template = '''
heat_template_version: 2016-04-08
description: This template to define a neutron bandwidth limit rule.
resources:
  my_bandwidth_limit_rule:
    type: OS::Neutron::QoSBandwidthLimitRule
    properties:
      policy: 477e8273-60a7-4c41-b683-fdb0bc7cd151
      max_kbps: 1000
      max_burst_kbps: 1000
      tenant_id: d66c74c01d6c41b9846088c1ad9634d0
'''

dscp_marking_rule_template = '''
heat_template_version: 2016-04-08
description: This template to define a neutron DSCP marking rule.
resources:
  my_dscp_marking_rule:
    type: OS::Neutron::QoSDscpMarkingRule
    properties:
      policy: 477e8273-60a7-4c41-b683-fdb0bc7cd151
      dscp_mark: 16
      tenant_id: d66c74c01d6c41b9846088c1ad9634d0
'''

minimum_bandwidth_rule_template = '''
heat_template_version: 2016-04-08
description: This template to define a neutron minimum bandwidth rule.
resources:
  my_minimum_bandwidth_rule:
    type: OS::Neutron::QoSMinimumBandwidthRule
    properties:
      policy: 477e8273-60a7-4c41-b683-fdb0bc7cd151
      min_kbps: 1000
      tenant_id: d66c74c01d6c41b9846088c1ad9634d0
'''


class NeutronQoSPolicyTest(common.HeatTestCase):
    def setUp(self):
        super(NeutronQoSPolicyTest, self).setUp()

        self.ctx = utils.dummy_context()
        tpl = template_format.parse(qos_policy_template)
        self.stack = stack.Stack(
            self.ctx,
            'neutron_qos_policy_test',
            template.Template(tpl)
        )

        self.neutronclient = mock.MagicMock()
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)
        self.my_qos_policy = self.stack['my_qos_policy']
        self.my_qos_policy.client = mock.MagicMock(
            return_value=self.neutronclient)
        self.patchobject(self.my_qos_policy, 'physical_resource_name',
                         return_value='test_policy')

    def test_qos_policy_handle_create(self):
        policy = {
            'policy': {
                'description': 'a policy for test',
                'id': '9c1eb3fe-7bba-479d-bd43-1d497e53c384',
                'rules': [],
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0',
                'shared': True
            }
        }
        create_props = {'name': 'test_policy',
                        'description': 'a policy for test',
                        'shared': True,
                        'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'}

        self.neutronclient.create_qos_policy.return_value = policy
        self.my_qos_policy.handle_create()
        self.assertEqual('9c1eb3fe-7bba-479d-bd43-1d497e53c384',
                         self.my_qos_policy.resource_id)
        self.neutronclient.create_qos_policy.assert_called_once_with(
            {'policy': create_props}
        )

    def test_qos_policy_handle_delete(self):
        policy_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.my_qos_policy.resource_id = policy_id
        self.neutronclient.delete_qos_policy.return_value = None

        self.assertIsNone(self.my_qos_policy.handle_delete())
        self.neutronclient.delete_qos_policy.assert_called_once_with(
            self.my_qos_policy.resource_id)

    def test_qos_policy_handle_delete_not_found(self):
        policy_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.my_qos_policy.resource_id = policy_id
        not_found = self.neutronclient.NotFound
        self.neutronclient.delete_qos_policy.side_effect = not_found

        self.assertIsNone(self.my_qos_policy.handle_delete())
        self.neutronclient.delete_qos_policy.assert_called_once_with(
            self.my_qos_policy.resource_id)

    def test_qos_policy_handle_delete_resource_id_is_none(self):
        self.my_qos_policy.resource_id = None
        self.assertIsNone(self.my_qos_policy.handle_delete())
        self.assertEqual(0, self.neutronclient.delete_qos_policy.call_count)

    def test_qos_policy_handle_update(self):
        policy_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.my_qos_policy.resource_id = policy_id

        props = {
            'name': 'test_policy',
            'description': 'test',
            'shared': False
        }
        prop_dict = props.copy()
        update_snippet = rsrc_defn.ResourceDefinition(
            self.my_qos_policy.name,
            self.my_qos_policy.type(),
            props)

        # with name
        self.my_qos_policy.handle_update(json_snippet=update_snippet,
                                         tmpl_diff={},
                                         prop_diff=props)
        # without name
        props['name'] = None
        self.my_qos_policy.handle_update(json_snippet=update_snippet,
                                         tmpl_diff={},
                                         prop_diff=props)

        self.assertEqual(2, self.neutronclient.update_qos_policy.call_count)
        self.neutronclient.update_qos_policy.assert_called_with(
            policy_id, {'policy': prop_dict})

    def test_qos_policy_get_attr(self):
        self.my_qos_policy.resource_id = 'test policy'
        policy = {
            'policy': {
                'name': 'test_policy',
                'description': 'a policy for test',
                'id': '9c1eb3fe-7bba-479d-bd43-1d497e53c384',
                'rules': [],
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0',
                'shared': True
            }
        }
        self.neutronclient.show_qos_policy.return_value = policy

        self.assertEqual([], self.my_qos_policy.FnGetAtt('rules'))
        self.assertEqual(policy['policy'],
                         self.my_qos_policy.FnGetAtt('show'))
        self.neutronclient.show_qos_policy.assert_has_calls(
            [mock.call(self.my_qos_policy.resource_id)] * 2)


class NeutronQoSBandwidthLimitRuleTest(common.HeatTestCase):
    def setUp(self):
        super(NeutronQoSBandwidthLimitRuleTest, self).setUp()

        self.ctx = utils.dummy_context()
        tpl = template_format.parse(bandwidth_limit_rule_template)
        self.stack = stack.Stack(
            self.ctx,
            'neutron_bandwidth_limit_rule_test',
            template.Template(tpl)
        )

        self.neutronclient = mock.MagicMock()
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)
        self.bandwidth_limit_rule = self.stack['my_bandwidth_limit_rule']
        self.bandwidth_limit_rule.client = mock.MagicMock(
            return_value=self.neutronclient)
        self.find_mock = self.patchobject(
            neutron.neutronV20,
            'find_resourceid_by_name_or_id')
        self.policy_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.find_mock.return_value = self.policy_id

    def test_rule_handle_create(self):
        rule = {
            'bandwidth_limit_rule': {
                'id': 'cf0eab12-ef8b-4a62-98d0-70576583c17a',
                'max_kbps': 1000,
                'max_burst_kbps': 1000,
                'direction': 'egress',
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'
            }
        }

        create_props = {'max_kbps': 1000,
                        'max_burst_kbps': 1000,
                        'direction': 'egress',
                        'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'}
        self.neutronclient.create_bandwidth_limit_rule.return_value = rule

        self.bandwidth_limit_rule.handle_create()
        self.assertEqual('cf0eab12-ef8b-4a62-98d0-70576583c17a',
                         self.bandwidth_limit_rule.resource_id)
        self.neutronclient.create_bandwidth_limit_rule.assert_called_once_with(
            self.policy_id,
            {'bandwidth_limit_rule': create_props})

    def test_rule_handle_delete(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.bandwidth_limit_rule.resource_id = rule_id
        self.neutronclient.delete_bandwidth_limit_rule.return_value = None

        self.assertIsNone(self.bandwidth_limit_rule.handle_delete())
        self.neutronclient.delete_bandwidth_limit_rule.assert_called_once_with(
            rule_id, self.policy_id)

    def test_rule_handle_delete_not_found(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.bandwidth_limit_rule.resource_id = rule_id
        not_found = self.neutronclient.NotFound
        self.neutronclient.delete_bandwidth_limit_rule.side_effect = not_found

        self.assertIsNone(self.bandwidth_limit_rule.handle_delete())
        self.neutronclient.delete_bandwidth_limit_rule.assert_called_once_with(
            rule_id, self.policy_id)

    def test_rule_handle_delete_resource_id_is_none(self):
        self.bandwidth_limit_rule.resource_id = None
        self.assertIsNone(self.bandwidth_limit_rule.handle_delete())
        self.assertEqual(0,
                         self.neutronclient.bandwidth_limit_rule.call_count)

    def test_rule_handle_update(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.bandwidth_limit_rule.resource_id = rule_id

        prop_diff = {
            'max_kbps': 500,
            'max_burst_kbps': 400
        }

        self.bandwidth_limit_rule.handle_update(
            json_snippet={},
            tmpl_diff={},
            prop_diff=prop_diff.copy())

        self.neutronclient.update_bandwidth_limit_rule.assert_called_once_with(
            rule_id,
            self.policy_id,
            {'bandwidth_limit_rule': prop_diff})

    def test_rule_get_attr(self):
        self.bandwidth_limit_rule.resource_id = 'test rule'
        rule = {
            'bandwidth_limit_rule': {
                'id': 'cf0eab12-ef8b-4a62-98d0-70576583c17a',
                'max_kbps': 1000,
                'max_burst_kbps': 1000,
                'direction': 'egress',
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'
            }
        }
        self.neutronclient.show_bandwidth_limit_rule.return_value = rule

        self.assertEqual(rule['bandwidth_limit_rule'],
                         self.bandwidth_limit_rule.FnGetAtt('show'))

        self.neutronclient.show_bandwidth_limit_rule.assert_called_once_with(
            self.bandwidth_limit_rule.resource_id, self.policy_id)


class NeutronQoSDscpMarkingRuleTest(common.HeatTestCase):
    def setUp(self):
        super(NeutronQoSDscpMarkingRuleTest, self).setUp()

        self.ctx = utils.dummy_context()
        tpl = template_format.parse(dscp_marking_rule_template)
        self.stack = stack.Stack(
            self.ctx,
            'neutron_dscp_marking_rule_test',
            template.Template(tpl)
        )

        self.neutronclient = mock.MagicMock()
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)
        self.dscp_marking_rule = self.stack['my_dscp_marking_rule']
        self.dscp_marking_rule.client = mock.MagicMock(
            return_value=self.neutronclient)
        self.find_mock = self.patchobject(
            neutron.neutronV20,
            'find_resourceid_by_name_or_id')
        self.policy_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.find_mock.return_value = self.policy_id

    def test_rule_handle_create(self):
        rule = {
            'dscp_marking_rule': {
                'id': 'cf0eab12-ef8b-4a62-98d0-70576583c17a',
                'dscp_mark': 16,
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'
            }
        }

        create_props = {'dscp_mark': 16,
                        'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'}
        self.neutronclient.create_dscp_marking_rule.return_value = rule

        self.dscp_marking_rule.handle_create()
        self.assertEqual('cf0eab12-ef8b-4a62-98d0-70576583c17a',
                         self.dscp_marking_rule.resource_id)
        self.neutronclient.create_dscp_marking_rule.assert_called_once_with(
            self.policy_id,
            {'dscp_marking_rule': create_props})

    def test_rule_handle_delete(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.dscp_marking_rule.resource_id = rule_id
        self.neutronclient.delete_dscp_marking_rule.return_value = None

        self.assertIsNone(self.dscp_marking_rule.handle_delete())
        self.neutronclient.delete_dscp_marking_rule.assert_called_once_with(
            rule_id, self.policy_id)

    def test_rule_handle_delete_not_found(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.dscp_marking_rule.resource_id = rule_id
        not_found = self.neutronclient.NotFound
        self.neutronclient.delete_dscp_marking_rule.side_effect = not_found

        self.assertIsNone(self.dscp_marking_rule.handle_delete())
        self.neutronclient.delete_dscp_marking_rule.assert_called_once_with(
            rule_id, self.policy_id)

    def test_rule_handle_delete_resource_id_is_none(self):
        self.dscp_marking_rule.resource_id = None
        self.assertIsNone(self.dscp_marking_rule.handle_delete())
        self.assertEqual(0,
                         self.neutronclient.dscp_marking_rule.call_count)

    def test_rule_handle_update(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.dscp_marking_rule.resource_id = rule_id

        prop_diff = {
            'dscp_mark': 8
        }

        self.dscp_marking_rule.handle_update(
            json_snippet={},
            tmpl_diff={},
            prop_diff=prop_diff)

        self.neutronclient.update_dscp_marking_rule.assert_called_once_with(
            rule_id,
            self.policy_id,
            {'dscp_marking_rule': prop_diff})

    def test_rule_get_attr(self):
        self.dscp_marking_rule.resource_id = 'test rule'
        rule = {
            'dscp_marking_rule': {
                'id': 'cf0eab12-ef8b-4a62-98d0-70576583c17a',
                'dscp_mark': 8,
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'
            }
        }
        self.neutronclient.show_dscp_marking_rule.return_value = rule

        self.assertEqual(rule['dscp_marking_rule'],
                         self.dscp_marking_rule.FnGetAtt('show'))

        self.neutronclient.show_dscp_marking_rule.assert_called_once_with(
            self.dscp_marking_rule.resource_id, self.policy_id)


class NeutronQoSMinimumBandwidthRuleTest(common.HeatTestCase):
    def setUp(self):
        super(NeutronQoSMinimumBandwidthRuleTest, self).setUp()

        self.ctx = utils.dummy_context()
        tpl = template_format.parse(minimum_bandwidth_rule_template)
        self.stack = stack.Stack(
            self.ctx,
            'neutron_minimum_bandwidth_rule_test',
            template.Template(tpl)
        )

        self.neutronclient = mock.MagicMock()
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)
        self.minimum_bandwidth_rule = self.stack['my_minimum_bandwidth_rule']
        self.minimum_bandwidth_rule.client = mock.MagicMock(
            return_value=self.neutronclient)
        self.find_mock = self.patchobject(
            neutron.neutronV20,
            'find_resourceid_by_name_or_id')
        self.policy_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.find_mock.return_value = self.policy_id

    def test_rule_handle_create(self):
        rule = {
            'minimum_bandwidth_rule': {
                'id': 'cf0eab12-ef8b-4a62-98d0-70576583c17a',
                'min_kbps': 1000,
                'direction': 'egress',
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'
            }
        }

        create_props = {'min_kbps': 1000,
                        'direction': 'egress',
                        'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'}
        self.neutronclient.create_minimum_bandwidth_rule.return_value = rule

        self.minimum_bandwidth_rule.handle_create()
        self.assertEqual('cf0eab12-ef8b-4a62-98d0-70576583c17a',
                         self.minimum_bandwidth_rule.resource_id)
        self.neutronclient.create_minimum_bandwidth_rule.\
            assert_called_once_with(
                self.policy_id,
                {'minimum_bandwidth_rule': create_props})

    def test_rule_handle_delete(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.minimum_bandwidth_rule.resource_id = rule_id
        self.neutronclient.delete_minimum_bandwidth_rule.return_value = None

        self.assertIsNone(self.minimum_bandwidth_rule.handle_delete())
        self.neutronclient.delete_minimum_bandwidth_rule.\
            assert_called_once_with(rule_id, self.policy_id)

    def test_rule_handle_delete_not_found(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.minimum_bandwidth_rule.resource_id = rule_id
        not_found = self.neutronclient.NotFound
        self.neutronclient.delete_minimum_bandwidth_rule.side_effect =\
            not_found

        self.assertIsNone(self.minimum_bandwidth_rule.handle_delete())
        self.neutronclient.delete_minimum_bandwidth_rule.\
            assert_called_once_with(rule_id, self.policy_id)

    def test_rule_handle_delete_resource_id_is_none(self):
        self.minimum_bandwidth_rule.resource_id = None
        self.assertIsNone(self.minimum_bandwidth_rule.handle_delete())
        self.assertEqual(0,
                         self.neutronclient.minimum_bandwidth_rule.call_count)

    def test_rule_handle_update(self):
        rule_id = 'cf0eab12-ef8b-4a62-98d0-70576583c17a'
        self.minimum_bandwidth_rule.resource_id = rule_id

        prop_diff = {
            'min_kbps': 500
        }

        self.minimum_bandwidth_rule.handle_update(
            json_snippet={},
            tmpl_diff={},
            prop_diff=prop_diff.copy())

        self.neutronclient.update_minimum_bandwidth_rule.\
            assert_called_once_with(
                rule_id,
                self.policy_id,
                {'minimum_bandwidth_rule': prop_diff})

    def test_rule_get_attr(self):
        self.minimum_bandwidth_rule.resource_id = 'test rule'
        rule = {
            'minimum_bandwidth_rule': {
                'id': 'cf0eab12-ef8b-4a62-98d0-70576583c17a',
                'min_kbps': 1000,
                'direction': 'egress',
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0'
            }
        }
        self.neutronclient.show_minimum_bandwidth_rule.return_value = rule

        self.assertEqual(rule['minimum_bandwidth_rule'],
                         self.minimum_bandwidth_rule.FnGetAtt('show'))

        self.neutronclient.show_minimum_bandwidth_rule.assert_called_once_with(
            self.minimum_bandwidth_rule.resource_id, self.policy_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import identifier
from heat.objects import event as event_object


class Event(object):
    """Class representing a Resource state change."""

    def __init__(self, context, stack, action, status, reason,
                 physical_resource_id, resource_prop_data_id,
                 resource_properties, resource_name,
                 resource_type, uuid=None, timestamp=None, id=None):
        """Initialise from a context, stack, and event information.

        The timestamp and database ID may also be initialised if the event is
        already in the database.
        """
        self.context = context
        self._stack_identifier = stack.identifier()
        self.action = action
        self.status = status
        self.reason = reason
        self.physical_resource_id = physical_resource_id
        self.resource_name = resource_name
        self.resource_type = resource_type
        self.rsrc_prop_data_id = resource_prop_data_id
        self.resource_properties = resource_properties
        if self.resource_properties is None:
            self.resource_properties = {}
        self.uuid = uuid
        self.timestamp = timestamp
        self.id = id

    def store(self):
        """Store the Event in the database."""
        ev = {
            'resource_name': self.resource_name,
            'physical_resource_id': self.physical_resource_id,
            'stack_id': self._stack_identifier.stack_id,
            'resource_action': self.action,
            'resource_status': self.status,
            'resource_status_reason': self.reason,
            'resource_type': self.resource_type,
        }

        if self.uuid is not None:
            ev['uuid'] = self.uuid

        if self.timestamp is not None:
            ev['created_at'] = self.timestamp

        if self.rsrc_prop_data_id is not None:
            ev['rsrc_prop_data_id'] = self.rsrc_prop_data_id

        new_ev = event_object.Event.create(self.context, ev)

        self.id = new_ev.id
        self.timestamp = new_ev.created_at
        self.uuid = new_ev.uuid
        return self.id

    def identifier(self):
        """Return a unique identifier for the event."""
        if self.uuid is None:
            return None

        res_id = identifier.ResourceIdentifier(
            resource_name=self.resource_name, **self._stack_identifier)

        return identifier.EventIdentifier(event_id=str(self.uuid), **res_id)

    def as_dict(self):
        return {
            'timestamp': self.timestamp.isoformat(),
            'version': '0.1',
            'type': 'os.heat.event',
            'id': self.uuid,
            'payload': {
                'resource_name': self.resource_name,
                'physical_resource_id': self.physical_resource_id,
                'stack_id': self._stack_identifier.stack_id,
                'resource_action': self.action,
                'resource_status': self.status,
                'resource_status_reason': self.reason,
                'resource_type': self.resource_type,
                'resource_properties': self.resource_properties,
                'version': '0.1'
            }
        }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from oslo_serialization import jsonutils

from heat.common import exception
from heat.common.i18n import _
from heat.common import param_utils
from heat.engine import constraints as constr
from heat.engine import function
from heat.engine.hot import parameters as hot_param
from heat.engine import parameters
from heat.engine import support
from heat.engine import translation as trans

SCHEMA_KEYS = (
    REQUIRED, IMPLEMENTED, DEFAULT, TYPE, SCHEMA,
    ALLOWED_PATTERN, MIN_VALUE, MAX_VALUE, ALLOWED_VALUES,
    MIN_LENGTH, MAX_LENGTH, DESCRIPTION, UPDATE_ALLOWED,
    IMMUTABLE,
) = (
    'Required', 'Implemented', 'Default', 'Type', 'Schema',
    'AllowedPattern', 'MinValue', 'MaxValue', 'AllowedValues',
    'MinLength', 'MaxLength', 'Description', 'UpdateAllowed',
    'Immutable',
)


class Schema(constr.Schema):
    """Schema class for validating resource properties.

    This class is used for defining schema constraints for resource properties.
    It inherits generic validation features from the base Schema class and add
    processing that is specific to resource properties.
    """

    KEYS = (
        TYPE, DESCRIPTION, DEFAULT, SCHEMA, REQUIRED, CONSTRAINTS,
        UPDATE_ALLOWED, IMMUTABLE,
    ) = (
        'type', 'description', 'default', 'schema', 'required', 'constraints',
        'update_allowed', 'immutable',
    )

    def __init__(self, data_type, description=None,
                 default=None, schema=None,
                 required=False, constraints=None,
                 implemented=True,
                 update_allowed=False,
                 immutable=False,
                 support_status=support.SupportStatus(),
                 allow_conversion=False):
        super(Schema, self).__init__(data_type, description, default,
                                     schema, required, constraints,
                                     immutable=immutable)
        self.implemented = implemented
        self.update_allowed = update_allowed
        self.support_status = support_status
        self.allow_conversion = allow_conversion
        # validate structural correctness of schema itself
        self.validate()

    def validate(self, context=None):
        super(Schema, self).validate()
        # check that update_allowed and immutable
        # do not contradict each other
        if self.update_allowed and self.immutable:
            msg = _("Options %(ua)s and %(im)s "
                    "cannot both be True") % {
                        'ua': UPDATE_ALLOWED,
                        'im': IMMUTABLE}
            raise exception.InvalidSchemaError(message=msg)

    @classmethod
    def from_legacy(cls, schema_dict):
        """Return a Property Schema object from a legacy schema dictionary."""

        # Check for fully-fledged Schema objects
        if isinstance(schema_dict, cls):
            return schema_dict

        unknown = [k for k in schema_dict if k not in SCHEMA_KEYS]
        if unknown:
            raise exception.InvalidSchemaError(
                message=_('Unknown key(s) %s') % unknown)

        def constraints():
            def get_num(key):
                val = schema_dict.get(key)
                if val is not None:
                    val = Schema.str_to_num(val)
                return val

            if MIN_VALUE in schema_dict or MAX_VALUE in schema_dict:
                yield constr.Range(get_num(MIN_VALUE), get_num(MAX_VALUE))
            if MIN_LENGTH in schema_dict or MAX_LENGTH in schema_dict:
                yield constr.Length(get_num(MIN_LENGTH), get_num(MAX_LENGTH))
            if ALLOWED_VALUES in schema_dict:
                yield constr.AllowedValues(schema_dict[ALLOWED_VALUES])
            if ALLOWED_PATTERN in schema_dict:
                yield constr.AllowedPattern(schema_dict[ALLOWED_PATTERN])

        try:
            data_type = schema_dict[TYPE]
        except KeyError:
            raise exception.InvalidSchemaError(
                message=_('No %s specified') % TYPE)

        if SCHEMA in schema_dict:
            if data_type == Schema.LIST:
                ss = cls.from_legacy(schema_dict[SCHEMA])
            elif data_type == Schema.MAP:
                schema_dicts = schema_dict[SCHEMA].items()
                ss = dict((n, cls.from_legacy(sd)) for n, sd in schema_dicts)
            else:
                raise exception.InvalidSchemaError(
                    message=_('%(schema)s supplied for %(type)s %(data)s') %
                    dict(schema=SCHEMA, type=TYPE, data=data_type))
        else:
            ss = None

        return cls(data_type,
                   description=schema_dict.get(DESCRIPTION),
                   default=schema_dict.get(DEFAULT),
                   schema=ss,
                   required=schema_dict.get(REQUIRED, False),
                   constraints=list(constraints()),
                   implemented=schema_dict.get(IMPLEMENTED, True),
                   update_allowed=schema_dict.get(UPDATE_ALLOWED, False),
                   immutable=schema_dict.get(IMMUTABLE, False))

    @classmethod
    def from_parameter(cls, param):
        """Return a Property Schema corresponding to a Parameter Schema.

        Convert a parameter schema from a provider template to a property
        Schema for the corresponding resource facade.
        """

        # map param types to property types
        param_type_map = {
            param.STRING: cls.STRING,
            param.NUMBER: cls.NUMBER,
            param.LIST: cls.LIST,
            param.MAP: cls.MAP,
            param.BOOLEAN: cls.BOOLEAN
        }

        # allow_conversion allows slightly more flexible type conversion
        # where property->parameter types don't align, primarily when
        # a json parameter value is passed via a Map property, which requires
        # some coercion to pass strings or lists (which are both valid for
        # Json parameters but not for Map properties).
        allow_conversion = (param.type == param.MAP
                            or param.type == param.LIST)

        # make update_allowed true by default on TemplateResources
        # as the template should deal with this.
        return cls(data_type=param_type_map.get(param.type, cls.MAP),
                   description=param.description,
                   required=param.required,
                   constraints=param.constraints,
                   update_allowed=True,
                   immutable=False,
                   allow_conversion=allow_conversion)

    def allowed_param_prop_type(self):
        """Return allowed type of Property Schema converted from parameter.

        Especially, when generating Schema from parameter, Integer Property
        Schema will be supplied by Number parameter.
        """
        param_type_map = {
            self.INTEGER: self.NUMBER,
            self.STRING: self.STRING,
            self.NUMBER: self.NUMBER,
            self.BOOLEAN: self.BOOLEAN,
            self.LIST: self.LIST,
            self.MAP: self.MAP
        }

        return param_type_map[self.type]

    def __getitem__(self, key):
        if key == self.UPDATE_ALLOWED:
            return self.update_allowed
        elif key == self.IMMUTABLE:
            return self.immutable
        else:
            return super(Schema, self).__getitem__(key)


def schemata(schema_dicts):
    """Return dictionary of Schema objects for given dictionary of schemata.

    The input schemata are converted from the legacy (dictionary-based)
    format to Schema objects where necessary.
    """
    return dict((n, Schema.from_legacy(s)) for n, s in schema_dicts.items())


class Property(object):

    def __init__(self, schema, name=None, context=None, path=None):
        self.schema = Schema.from_legacy(schema)
        self.name = name
        self.context = context
        self.path = self.make_path(name, path)

    def required(self):
        return self.schema.required

    def implemented(self):
        return self.schema.implemented

    def update_allowed(self):
        return self.schema.update_allowed

    def immutable(self):
        return self.schema.immutable

    def has_default(self):
        return self.schema.default is not None

    def default(self):
        return self.schema.default

    def type(self):
        return self.schema.type

    def support_status(self):
        return self.schema.support_status

    def make_path(self, name, path=None):
        if path is None:
            path = ''
        if name is None:
            name = ''

        if isinstance(name, int) or name.isdigit():
            name = str(name)

        delim = '' if not path or path.endswith('.') else '.'
        return delim.join([path, name])

    def _get_integer(self, value):
        if value is None:
            value = self.has_default() and self.default() or 0
        try:
            value = int(value)
        except ValueError:
            raise TypeError(_("Value '%s' is not an integer") % value)
        else:
            return value

    def _get_number(self, value):
        if value is None:
            value = self.has_default() and self.default() or 0
        return Schema.str_to_num(value)

    def _get_string(self, value):
        if value is None:
            value = self.has_default() and self.default() or ''
        if not isinstance(value, str):
            if isinstance(value, (bool, int)):
                value = str(value)
            else:
                raise ValueError(_('Value must be a string; got %r') % value)
        return value

    def _get_children(self, child_values, keys=None, validate=False,
                      translation=None):
        if self.schema.schema is not None:
            if keys is None:
                keys = list(self.schema.schema)
            schemata = dict((k, self.schema.schema[k]) for k in keys)
            properties = Properties(schemata, dict(child_values),
                                    context=self.context,
                                    parent_name=self.path,
                                    translation=translation)
            if validate:
                properties.validate()

            return ((k, properties[k]) for k in keys)
        else:
            return child_values

    def _get_map(self, value, validate=False, translation=None):
        if value is None:
            value = self.default() if self.has_default() else {}
        if not isinstance(value, collections.abc.Mapping):
            # This is to handle passing Lists via Json parameters exposed
            # via a provider resource, in particular lists-of-dicts which
            # cannot be handled correctly via comma_delimited_list
            if self.schema.allow_conversion:
                if isinstance(value, str):
                    return value
                elif isinstance(value, collections.abc.Sequence):
                    return jsonutils.dumps(value)
            raise TypeError(_('"%s" is not a map') % value)

        return dict(self._get_children(value.items(),
                                       validate=validate,
                                       translation=translation))

    def _get_list(self, value, validate=False, translation=None):
        if value is None:
            value = self.has_default() and self.default() or []
        if self.schema.allow_conversion and isinstance(value, str):
            value = param_utils.delim_string_to_list(value)
        if (not isinstance(value, collections.abc.Sequence) or
                isinstance(value, str)):
            raise TypeError(_('"%s" is not a list') % repr(value))

        return [v[1] for v in self._get_children(enumerate(value),
                                                 range(len(value)),
                                                 validate=validate,
                                                 translation=translation)]

    def _get_bool(self, value):
        """Get value for boolean property.

        Explicitly checking for bool, or string with lower value
        "true" or "false", to avoid integer values.
        """
        if value is None:
            value = self.has_default() and self.default() or False
        if isinstance(value, bool):
            return value
        if isinstance(value, str):
            normalised = value.lower()
            if normalised not in ['true', 'false']:
                raise ValueError(_('"%s" is not a valid boolean') % normalised)
            return normalised == 'true'

        raise TypeError(_('"%s" is not a valid boolean') % value)

    def get_value(self, value, validate=False, translation=None):
        """Get value from raw value and sanitize according to data type."""

        t = self.type()
        if t == Schema.STRING:
            _value = self._get_string(value)
        elif t == Schema.INTEGER:
            _value = self._get_integer(value)
        elif t == Schema.NUMBER:
            _value = self._get_number(value)
        elif t == Schema.MAP:
            _value = self._get_map(value, validate, translation)
        elif t == Schema.LIST:
            _value = self._get_list(value, validate, translation)
        elif t == Schema.BOOLEAN:
            _value = self._get_bool(value)
        elif t == Schema.ANY:
            _value = value

        if validate:
            self.schema.validate_constraints(_value, self.context)

        return _value


def _default_resolver(d, nullable=False):
    return d


class Properties(collections.abc.Mapping):

    def __init__(self, schema, data, resolver=_default_resolver,
                 parent_name=None,
                 context=None, section=None, translation=None,
                 rsrc_description=None):
        self.props = dict((k, Property(s, k, context, path=parent_name))
                          for k, s in schema.items())
        self.resolve = resolver
        self.data = data
        self.error_prefix = [section] if section is not None else []
        self.parent_name = parent_name
        self.context = context
        self.translation = (trans.Translation(properties=self)
                            if translation is None else translation)
        self.rsrc_description = rsrc_description or None

    def update_translation(self, rules, client_resolve=True,
                           ignore_resolve_error=False):
        self.translation.set_rules(rules, client_resolve=client_resolve,
                                   ignore_resolve_error=ignore_resolve_error)

    @staticmethod
    def schema_from_params(params_snippet):
        """Create properties schema from the parameters section of a template.

        :param params_snippet: parameter definition from a template
        :returns: equivalent properties schemata for the specified parameters
        """
        if params_snippet:
            return dict((n, Schema.from_parameter(p)) for n, p
                        in params_snippet.items())
        return {}

    def validate(self, with_value=True):
        try:
            for key in self.data:
                if key not in self.props:
                    msg = _("Unknown Property %s") % key
                    raise exception.StackValidationFailed(message=msg)

            for (key, prop) in self.props.items():
                if (self.translation.is_deleted(prop.path) or
                        self.translation.is_replaced(prop.path)):
                    continue
                if with_value:
                    try:
                        self._get_property_value(key, validate=True)
                    except exception.StackValidationFailed as ex:
                        path = [key]
                        path.extend(ex.path)
                        raise exception.StackValidationFailed(
                            path=path, message=ex.error_message)
                    except ValueError as e:
                        if prop.required() and key not in self.data:
                            path = []
                        else:
                            path = [key]
                        raise exception.StackValidationFailed(
                            path=path, message=str(e))

                # are there unimplemented Properties
                if not prop.implemented() and key in self.data:
                    msg = _("Property %s not implemented yet") % key
                    raise exception.StackValidationFailed(message=msg)
        except exception.StackValidationFailed as ex:
            # NOTE(prazumovsky): should reraise exception for adding specific
            # error name and error_prefix to path for correct error message
            # building.
            path = self.error_prefix
            path.extend(ex.path)
            raise exception.StackValidationFailed(
                error=ex.error or 'Property error',
                path=path,
                message=ex.error_message
            )

    def _find_deps_any_in_init(self, unresolved_value):
        deps = function.dependencies(unresolved_value)
        if any(res.action == res.INIT for res in deps):
            return True

    def get_user_value(self, key):
        if key not in self:
            raise KeyError(_('Invalid Property %s') % key)

        prop = self.props[key]
        value, found = self._resolve_user_value(key, prop, validate=False)
        return value

    def _resolve_user_value(self, key, prop, validate):
        """Return the user-supplied value (or None), and whether it was found.

        This allows us to distinguish between, on the one hand, either a
        Function that returns None or an explicit null value passed and, on the
        other hand, either no value passed or a Macro that returns Ellipsis,
        meaning that the result should be treated the same as if no value were
        passed.
        """
        if key not in self.data:
            return None, False

        if (self.translation.is_deleted(prop.path) or
                self.translation.is_replaced(prop.path)):
            return None, False

        try:
            unresolved_value = self.data[key]
            if validate:
                if self._find_deps_any_in_init(unresolved_value):
                    validate = False

            value = self.resolve(unresolved_value, nullable=True)
            if value is Ellipsis:
                # Treat as if the property value were not specified at all
                return None, False

            if self.translation.has_translation(prop.path):
                value = self.translation.translate(prop.path,
                                                   value,
                                                   self.data)

            return prop.get_value(value, validate,
                                  translation=self.translation), True
        # Children can raise StackValidationFailed with unique path which
        # is necessary for further use in StackValidationFailed exception.
        # So we need to handle this exception in this method.
        except exception.StackValidationFailed as e:
            raise exception.StackValidationFailed(path=e.path,
                                                  message=e.error_message)
        # the resolver function could raise any number of exceptions,
        # so handle this generically
        except Exception as e:
            raise ValueError(str(e))

    def _get_property_value(self, key, validate=False):
        if key not in self:
            raise KeyError(_('Invalid Property %s') % key)

        prop = self.props[key]
        value, found = self._resolve_user_value(key, prop, validate)
        if found:
            return value
        if self.translation.has_translation(prop.path):
            value = self.translation.translate(prop.path, prop_data=self.data,
                                               validate=validate)
            if value is not None or prop.has_default():
                return prop.get_value(value)

        if prop.has_default():
            return prop.get_value(None, validate,
                                  translation=self.translation)
        elif prop.required():
            raise ValueError(_('Property %s not assigned') % key)
        elif key == 'description' and prop.schema.update_allowed:
            return self.rsrc_description
        else:
            return None

    def __getitem__(self, key):
        return self._get_property_value(key)

    def __len__(self):
        return len(self.props)

    def __contains__(self, key):
        return key in self.props

    def __iter__(self):
        return iter(self.props)

    @staticmethod
    def _param_def_from_prop(schema):
        """Return a template parameter definition corresponding to property."""
        param_type_map = {
            schema.INTEGER: parameters.Schema.NUMBER,
            schema.STRING: parameters.Schema.STRING,
            schema.NUMBER: parameters.Schema.NUMBER,
            schema.BOOLEAN: parameters.Schema.BOOLEAN,
            schema.MAP: parameters.Schema.MAP,
            schema.LIST: parameters.Schema.LIST,
        }

        def param_items():
            yield parameters.TYPE, param_type_map[schema.type]

            if schema.description is not None:
                yield parameters.DESCRIPTION, schema.description

            if schema.default is not None:
                yield parameters.DEFAULT, schema.default

            for constraint in schema.constraints:
                if isinstance(constraint, constr.Length):
                    if constraint.min is not None:
                        yield parameters.MIN_LENGTH, constraint.min
                    if constraint.max is not None:
                        yield parameters.MAX_LENGTH, constraint.max
                elif isinstance(constraint, constr.Range):
                    if constraint.min is not None:
                        yield parameters.MIN_VALUE, constraint.min
                    if constraint.max is not None:
                        yield parameters.MAX_VALUE, constraint.max
                elif isinstance(constraint, constr.AllowedValues):
                    yield parameters.ALLOWED_VALUES, list(constraint.allowed)
                elif isinstance(constraint, constr.AllowedPattern):
                    yield parameters.ALLOWED_PATTERN, constraint.pattern

            if schema.type == schema.BOOLEAN:
                yield parameters.ALLOWED_VALUES, ['True', 'true',
                                                  'False', 'false']

        return dict(param_items())

    @staticmethod
    def _prop_def_from_prop(name, schema):
        """Return a provider template property definition for a property."""
        if schema.type == Schema.LIST:
            return {'Fn::Split': [',', {'Ref': name}]}
        else:
            return {'Ref': name}

    @staticmethod
    def _hot_param_def_from_prop(schema):
        """Parameter definition corresponding to property for hot template."""
        param_type_map = {
            schema.INTEGER: hot_param.HOTParamSchema.NUMBER,
            schema.STRING: hot_param.HOTParamSchema.STRING,
            schema.NUMBER: hot_param.HOTParamSchema.NUMBER,
            schema.BOOLEAN: hot_param.HOTParamSchema.BOOLEAN,
            schema.MAP: hot_param.HOTParamSchema.MAP,
            schema.LIST: hot_param.HOTParamSchema.LIST,
        }

        def param_items():
            yield hot_param.HOTParamSchema.TYPE, param_type_map[schema.type]

            if schema.description is not None:
                yield hot_param.HOTParamSchema.DESCRIPTION, schema.description

            if schema.default is not None:
                yield hot_param.HOTParamSchema.DEFAULT, schema.default

            def constraint_items(constraint):

                def range_min_max(constraint):
                    if constraint.min is not None:
                        yield hot_param.MIN, constraint.min
                    if constraint.max is not None:
                        yield hot_param.MAX, constraint.max

                if isinstance(constraint, constr.Length):
                    yield hot_param.LENGTH, dict(range_min_max(constraint))
                elif isinstance(constraint, constr.Range):
                    yield hot_param.RANGE, dict(range_min_max(constraint))
                elif isinstance(constraint, constr.AllowedValues):
                    yield hot_param.ALLOWED_VALUES, list(constraint.allowed)
                elif isinstance(constraint, constr.AllowedPattern):
                    yield hot_param.ALLOWED_PATTERN, constraint.pattern

            if schema.constraints:
                yield (hot_param.HOTParamSchema.CONSTRAINTS,
                       [dict(constraint_items(constraint)) for constraint
                        in schema.constraints])

        return dict(param_items())

    @staticmethod
    def _hot_prop_def_from_prop(name, schema):
        """Return a provider template property definition for a property."""
        return {'get_param': name}

    @classmethod
    def schema_to_parameters_and_properties(cls, schema, template_type='cfn'):
        """Convert a schema to template parameters and properties.

        This can be used to generate a provider template that matches the
        given properties schemata.

        :param schema: A resource type's properties_schema
        :returns: A tuple of params and properties dicts

        ex: input:  {'foo': {'Type': 'List'}}
            output: {'foo': {'Type': 'CommaDelimitedList'}},
                    {'foo': {'Fn::Split': {'Ref': 'foo'}}}

        ex: input:  {'foo': {'Type': 'String'}, 'bar': {'Type': 'Map'}}
            output: {'foo': {'Type': 'String'}, 'bar': {'Type': 'Json'}},
                    {'foo': {'Ref': 'foo'}, 'bar': {'Ref': 'bar'}}
        """
        def param_prop_def_items(name, schema, template_type):
            if template_type == 'hot':
                param_def = cls._hot_param_def_from_prop(schema)
                prop_def = cls._hot_prop_def_from_prop(name, schema)
            else:
                param_def = cls._param_def_from_prop(schema)
                prop_def = cls._prop_def_from_prop(name, schema)
            return (name, param_def), (name, prop_def)

        if not schema:
            return {}, {}

        param_prop_defs = [param_prop_def_items(n, s, template_type)
                           for n, s in schemata(schema).items()
                           if s.implemented]
        param_items, prop_items = zip(*param_prop_defs)
        return dict(param_items), dict(prop_items)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for database migrations. This test case reads the configuration
file test_migrations.conf for database connection settings
to use in the tests. For each connection found in the config file,
the test case runs a series of test cases to ensure that migrations work
properly both upgrading and downgrading, and that no data loss occurs
if possible.
"""

import fixtures
import os

from migrate.versioning import repository
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures
from oslo_db.sqlalchemy import test_migrations
from oslo_db.sqlalchemy import utils
from oslotest import base as test_base
import sqlalchemy
import testtools

from heat.db.sqlalchemy import migrate_repo
from heat.db.sqlalchemy import migration
from heat.db.sqlalchemy import models
from heat.tests import common


class DBNotAllowed(Exception):
    pass


class BannedDBSchemaOperations(fixtures.Fixture):
    """Ban some operations for migrations"""
    def __init__(self, banned_resources=None):
        super(BannedDBSchemaOperations, self).__init__()
        self._banned_resources = banned_resources or []

    @staticmethod
    def _explode(resource, op):
        print('%s.%s()' % (resource, op))
        raise DBNotAllowed(
            'Operation %s.%s() is not allowed in a database migration' % (
                resource, op))

    def setUp(self):
        super(BannedDBSchemaOperations, self).setUp()
        for thing in self._banned_resources:
            self.useFixture(fixtures.MonkeyPatch(
                'sqlalchemy.%s.drop' % thing,
                lambda *a, **k: self._explode(thing, 'drop')))
            self.useFixture(fixtures.MonkeyPatch(
                'sqlalchemy.%s.alter' % thing,
                lambda *a, **k: self._explode(thing, 'alter')))


class TestBannedDBSchemaOperations(testtools.TestCase):
    def test_column(self):
        column = sqlalchemy.Column()
        with BannedDBSchemaOperations(['Column']):
            self.assertRaises(DBNotAllowed, column.drop)
            self.assertRaises(DBNotAllowed, column.alter)

    def test_table(self):
        table = sqlalchemy.Table()
        with BannedDBSchemaOperations(['Table']):
            self.assertRaises(DBNotAllowed, table.drop)
            self.assertRaises(DBNotAllowed, table.alter)


class HeatMigrationsCheckers(test_migrations.WalkVersionsMixin,
                             common.FakeLogMixin):
    """Test sqlalchemy-migrate migrations."""

    snake_walk = False
    downgrade = False

    @property
    def INIT_VERSION(self):
        return migration.INIT_VERSION

    @property
    def REPOSITORY(self):
        migrate_file = migrate_repo.__file__
        return repository.Repository(
            os.path.abspath(os.path.dirname(migrate_file))
        )

    @property
    def migration_api(self):
        temp = __import__('oslo_db.sqlalchemy.migration', globals(),
                          locals(), ['versioning_api'], 0)
        return temp.versioning_api

    @property
    def migrate_engine(self):
        return self.engine

    def migrate_up(self, version, with_data=False):
        """Check that migrations don't cause downtime.

        Schema migrations can be done online, allowing for rolling upgrades.
        """
        # NOTE(xek): This is a list of migrations where we allow dropping
        # things. The rules for adding exceptions are very very specific.
        # Chances are you don't meet the critera.
        # Reviewers: DO NOT ALLOW THINGS TO BE ADDED HERE
        exceptions = [
            64,  # drop constraint
            86,  # drop watch_rule/watch_data tables
        ]
        # Reviewers: DO NOT ALLOW THINGS TO BE ADDED HERE

        # NOTE(xek): We start requiring things be additive in
        # liberty, so ignore all migrations before that point.
        LIBERTY_START = 63

        if version >= LIBERTY_START and version not in exceptions:
            banned = ['Table', 'Column']
        else:
            banned = None
        with BannedDBSchemaOperations(banned):
            super(HeatMigrationsCheckers, self).migrate_up(version, with_data)

    def test_walk_versions(self):
        self.walk_versions(self.snake_walk, self.downgrade)

    def assertColumnExists(self, engine, table, column):
        t = utils.get_table(engine, table)
        self.assertIn(column, t.c)

    def assertColumnType(self, engine, table, column, sqltype):
        t = utils.get_table(engine, table)
        col = getattr(t.c, column)
        self.assertIsInstance(col.type, sqltype)

    def assertColumnNotExists(self, engine, table, column):
        t = utils.get_table(engine, table)
        self.assertNotIn(column, t.c)

    def assertColumnIsNullable(self, engine, table, column):
        t = utils.get_table(engine, table)
        col = getattr(t.c, column)
        self.assertTrue(col.nullable)

    def assertColumnIsNotNullable(self, engine, table, column_name):
        table = utils.get_table(engine, table)
        column = getattr(table.c, column_name)
        self.assertFalse(column.nullable)

    def assertIndexExists(self, engine, table, index):
        t = utils.get_table(engine, table)
        index_names = [idx.name for idx in t.indexes]
        self.assertIn(index, index_names)

    def assertIndexMembers(self, engine, table, index, members):
        self.assertIndexExists(engine, table, index)

        t = utils.get_table(engine, table)
        index_columns = []
        for idx in t.indexes:
            if idx.name == index:
                for ix in idx.columns:
                    index_columns.append(ix.name)
                break

        self.assertEqual(sorted(members), sorted(index_columns))

    def _check_073(self, engine, data):
        # check if column still exists and is not nullable.
        self.assertColumnIsNotNullable(engine, 'resource_data', 'resource_id')
        # Ensure that only one foreign key exists and is created as expected.
        inspector = sqlalchemy.engine.reflection.Inspector.from_engine(engine)
        resource_data_fkeys = inspector.get_foreign_keys('resource_data')
        self.assertEqual(1, len(resource_data_fkeys))
        fk = resource_data_fkeys[0]
        self.assertEqual('fk_resource_id', fk['name'])
        self.assertEqual(['resource_id'], fk['constrained_columns'])
        self.assertEqual('resource', fk['referred_table'])
        self.assertEqual(['id'], fk['referred_columns'])

    def _check_079(self, engine, data):
        self.assertColumnExists(engine, 'resource',
                                'rsrc_prop_data_id')
        self.assertColumnExists(engine, 'event',
                                'rsrc_prop_data_id')
        column_list = [('id', False),
                       ('data', True),
                       ('encrypted', True),
                       ('updated_at', True),
                       ('created_at', True)]

        for column in column_list:
            self.assertColumnExists(engine,
                                    'resource_properties_data', column[0])
            if not column[1]:
                self.assertColumnIsNotNullable(engine,
                                               'resource_properties_data',
                                               column[0])
            else:
                self.assertColumnIsNullable(engine,
                                            'resource_properties_data',
                                            column[0])

    def _check_080(self, engine, data):
        self.assertColumnExists(engine, 'resource',
                                'attr_data_id')


class DbTestCase(test_fixtures.OpportunisticDBTestMixin,
                 test_base.BaseTestCase):
    def setUp(self):
        super(DbTestCase, self).setUp()

        self.engine = enginefacade.writer.get_engine()
        self.sessionmaker = enginefacade.writer.get_sessionmaker()


class TestHeatMigrationsMySQL(DbTestCase, HeatMigrationsCheckers):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class TestHeatMigrationsPostgreSQL(DbTestCase, HeatMigrationsCheckers):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture


class TestHeatMigrationsSQLite(DbTestCase, HeatMigrationsCheckers):
    pass


class ModelsMigrationSyncMixin(object):

    def get_metadata(self):
        return models.BASE.metadata

    def get_engine(self):
        return self.engine

    def db_sync(self, engine):
        migration.db_sync(engine=engine)

    def include_object(self, object_, name, type_, reflected, compare_to):
        if name in ['migrate_version'] and type_ == 'table':
            return False
        return True


class ModelsMigrationsSyncMysql(DbTestCase,
                                ModelsMigrationSyncMixin,
                                test_migrations.ModelsMigrationsSync):
    FIXTURE = test_fixtures.MySQLOpportunisticFixture


class ModelsMigrationsSyncPostgres(DbTestCase,
                                   ModelsMigrationSyncMixin,
                                   test_migrations.ModelsMigrationsSync):
    FIXTURE = test_fixtures.PostgresqlOpportunisticFixture


class ModelsMigrationsSyncSQLite(DbTestCase,
                                 ModelsMigrationSyncMixin,
                                 test_migrations.ModelsMigrationsSync):
    pass



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import aodh
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.engine import stack as parser
from heat.engine import template as tmpl
from heat.tests import common
from heat.tests import utils


alarm_template = '''
heat_template_version: 2016-10-14
resources:
  cps_alarm:
    type: OS::Aodh::CompositeAlarm
    properties:
      description: test the composite alarm
      alarm_actions: []
      severity: moderate
      composite_rule:
        operator: or
        rules:
        - type: threshold
          meter_name: cpu_util
          evaluation_periods: 1
          period: 60
          statistic: avg
          threshold: 0.8
          comparison_operator: ge
          exclude_outliers: false
        - and:
          - type: threshold
            meter_name: disk.usage
            evaluation_periods: 1
            period: 60
            statistic: avg
            threshold: 0.8
            comparison_operator: ge
            exclude_outliers: false
          - type: threshold
            meter_name: mem_util
            evaluation_periods: 1
            period: 60
            statistic: avg
            threshold: 0.8
            comparison_operator: ge
            exclude_outliers: false
'''


FakeCompositeAlarm = {'other_attrs': 'val',
                      'alarm_id': 'foo'}


class CompositeAlarmTest(common.HeatTestCase):
    def setUp(self):
        super(CompositeAlarmTest, self).setUp()
        self.fa = mock.Mock()

    def create_stack(self, template=None):
        temp = template_format.parse(template)
        template = tmpl.Template(temp)
        ctx = utils.dummy_context()
        ctx.tenant = 'test_tenant'
        stack = parser.Stack(ctx, utils.random_name(), template,
                             disable_rollback=True)
        stack.store()

        self.patchobject(aodh.AodhClientPlugin,
                         '_create').return_value = self.fa

        self.patchobject(self.fa.alarm,
                         'create').return_value = FakeCompositeAlarm

        return stack

    def test_handle_create(self):
        """Test create the composite alarm."""

        test_stack = self.create_stack(template=alarm_template)
        test_stack.create()
        rsrc = test_stack['cps_alarm']
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

    def test_handle_update(self):
        """Test update the composite alarm."""

        test_stack = self.create_stack(template=alarm_template)
        update_mock = self.patchobject(self.fa.alarm, 'update')

        test_stack.create()
        rsrc = test_stack['cps_alarm']
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

        after_props = copy.deepcopy(rsrc.properties.data)
        update_props = {
            'enabled': False,
            'repeat_actions': False,
            'insufficient_data_actions': [],
            'ok_actions': ['signal_handler']
        }
        after_props.update(update_props)

        snippet = rsrc_defn.ResourceDefinition(rsrc.name,
                                               rsrc.type(),
                                               after_props)

        scheduler.TaskRunner(rsrc.update, snippet)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.assertEqual(1, update_mock.call_count)

    def test_validate(self):
        test_stack = self.create_stack(template=alarm_template)
        props = test_stack.t['resources']['cps_alarm']['Properties']
        props['composite_rule']['operator'] = 'invalid'
        res = test_stack['cps_alarm']
        error_msg = '"invalid" is not an allowed value'

        exc = self.assertRaises(exception.StackValidationFailed,
                                res.validate)
        self.assertIn(error_msg, str(exc))

    def test_show_resource(self):
        test_stack = self.create_stack(template=alarm_template)
        res = test_stack['cps_alarm']
        res.client().alarm.create.return_value = FakeCompositeAlarm
        res.client().alarm.get.return_value = FakeCompositeAlarm
        scheduler.TaskRunner(res.create)()
        self.assertEqual(FakeCompositeAlarm, res.FnGetAtt('show'))



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

from openstack.clustering.v1._proxy import Proxy
from openstack import exceptions
from oslo_config import cfg

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import senlin
from heat.engine.resources.openstack.senlin import policy
from heat.engine import scheduler
from heat.engine import template
from heat.tests import common
from heat.tests import utils


policy_stack_template = """
heat_template_version: 2016-04-08
description: Senlin Policy Template
resources:
  senlin-policy:
    type: OS::Senlin::Policy
    properties:
      name: SenlinPolicy
      type: senlin.policy.deletion-1.0
      properties:
        criteria: OLDEST_FIRST
      bindings:
        - cluster: c1
"""

policy_spec = {
    'type': 'senlin.policy.deletion',
    'version': '1.0',
    'properties': {
        'criteria': 'OLDEST_FIRST'
    }
}


class FakePolicy(object):
    def __init__(self, id='some_id', spec=None):
        self.id = id
        self.name = "SenlinPolicy"

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
        }


class SenlinPolicyTest(common.HeatTestCase):
    def setUp(self):
        super(SenlinPolicyTest, self).setUp()
        self.patchobject(senlin.ClusterConstraint, 'validate',
                         return_value=True)
        self.patchobject(senlin.PolicyTypeConstraint, 'validate',
                         return_value=True)
        self.senlin_mock = mock.MagicMock(spec=Proxy)
        self.senlin_mock.get_cluster.return_value = mock.Mock(
            id='c1_id')
        self.patchobject(policy.Policy, 'client',
                         return_value=self.senlin_mock)
        self.patchobject(senlin.SenlinClientPlugin, 'client',
                         return_value=self.senlin_mock)
        self.fake_p = FakePolicy()
        self.t = template_format.parse(policy_stack_template)

    def _init_policy(self, template):
        self.stack = utils.parse_stack(template)
        policy = self.stack['senlin-policy']
        return policy

    def _create_policy(self, template):
        policy = self._init_policy(template)
        self.senlin_mock.create_policy.return_value = self.fake_p
        self.senlin_mock.attach_policy_to_cluster.return_value = {
            'action': 'fake_action'}
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='SUCCEEDED')
        scheduler.TaskRunner(policy.create)()
        self.assertEqual((policy.CREATE, policy.COMPLETE),
                         policy.state)
        self.assertEqual(self.fake_p.id, policy.resource_id)
        self.senlin_mock.attach_policy_to_cluster.assert_called_once_with(
            'c1_id', policy.resource_id, enabled=True)
        self.senlin_mock.get_action.assert_called_once_with('fake_action')
        return policy

    def test_policy_create(self):
        self._create_policy(self.t)
        expect_kwargs = {
            'name': 'SenlinPolicy',
            'spec': policy_spec
        }
        self.senlin_mock.create_policy.assert_called_once_with(
            **expect_kwargs)

    def test_policy_create_fail(self):
        cfg.CONF.set_override('action_retry_limit', 0)
        policy = self._init_policy(self.t)
        self.senlin_mock.create_policy.return_value = self.fake_p
        self.senlin_mock.attach_policy_to_cluster.return_value = {
            'action': 'fake_action'}
        self.senlin_mock.get_action.return_value = mock.Mock(
            status='FAILED', status_reason='oops',
            action='CLUSTER_ATTACH_POLICY')
        create_task = scheduler.TaskRunner(policy.create)
        self.assertRaises(exception.ResourceFailure, create_task)
        self.assertEqual((policy.CREATE, policy.FAILED),
                         policy.state)
        err_msg = ('ResourceInError: resources.senlin-policy: Went to status '
                   'FAILED due to "Failed to execute CLUSTER_ATTACH_POLICY '
                   'for c1_id: oops"')
        self.assertEqual(err_msg, policy.status_reason)

    def test_policy_delete_not_found(self):
        self.senlin_mock.detach_policy_from_cluster.return_value = {
            'action': 'fake_action'}
        policy = self._create_policy(self.t)
        self.senlin_mock.get_policy.side_effect = [
            exceptions.ResourceNotFound('SenlinPolicy'),
        ]
        scheduler.TaskRunner(policy.delete)()
        self.senlin_mock.detach_policy_from_cluster.assert_called_once_with(
            'c1_id', policy.resource_id)
        self.senlin_mock.delete_policy.assert_called_once_with(
            policy.resource_id)

    def test_policy_delete_not_attached(self):
        policy = self._create_policy(self.t)
        self.senlin_mock.get_policy.side_effect = [
            exceptions.ResourceNotFound('SenlinPolicy'),
        ]
        self.senlin_mock.detach_policy_from_cluster.side_effect = [
            exceptions.HttpException(http_status=400),
        ]
        scheduler.TaskRunner(policy.delete)()
        self.senlin_mock.detach_policy_from_cluster.assert_called_once_with(
            'c1_id', policy.resource_id)
        self.senlin_mock.delete_policy.assert_called_once_with(
            policy.resource_id)

    def test_policy_update(self):
        policy = self._create_policy(self.t)
        # Mock translate rules
        self.senlin_mock.get_cluster.side_effect = [
            mock.Mock(id='c2_id'),
            mock.Mock(id='c1_id'),
            mock.Mock(id='c2_id'),
        ]
        new_t = copy.deepcopy(self.t)
        props = new_t['resources']['senlin-policy']['properties']
        props['bindings'] = [{'cluster': 'c2'}]
        props['name'] = 'new_name'
        rsrc_defns = template.Template(new_t).resource_definitions(self.stack)
        new_cluster = rsrc_defns['senlin-policy']
        self.senlin_mock.attach_policy_to_cluster.return_value = {
            'action': 'fake_action1'}
        self.senlin_mock.detach_policy_from_cluster.return_value = {
            'action': 'fake_action2'}
        self.senlin_mock.get_policy.return_value = self.fake_p
        scheduler.TaskRunner(policy.update, new_cluster)()
        self.assertEqual((policy.UPDATE, policy.COMPLETE), policy.state)
        self.senlin_mock.update_policy.assert_called_once_with(
            self.fake_p, name='new_name')
        self.senlin_mock.detach_policy_from_cluster.assert_called_once_with(
            'c1_id', policy.resource_id)
        self.senlin_mock.attach_policy_to_cluster.assert_called_with(
            'c2_id', policy.resource_id, enabled=True)

    def test_policy_resolve_attribute(self):
        excepted_show = {
            'id': 'some_id',
            'name': 'SenlinPolicy',
        }
        policy = self._create_policy(self.t)
        self.senlin_mock.get_policy.return_value = FakePolicy()
        self.assertEqual(excepted_show, policy._show_resource())



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy

from heat.common import exception
from heat.engine import function


# Field names that can be passed to Template.get_section_name() in order to
# determine the appropriate name for a particular template format.
FIELDS = (
    VALUE, DESCRIPTION,
) = (
    'Value', 'Description',
)


class OutputDefinition(object):
    """A definition of a stack output, independent of any template format."""

    def __init__(self, name, value, description=None):
        self.name = name
        self._value = value
        self._resolved_value = None
        self._description = description
        self._deps = None
        self._all_dep_attrs = None

    def validate(self):
        """Validate the output value without resolving it."""
        function.validate(self._value, VALUE)

    def required_resource_names(self):
        if self._deps is None:
            try:
                required_resources = function.dependencies(self._value)
                self._deps = set(map(lambda rp: rp.name, required_resources))
            except (exception.InvalidTemplateAttribute,
                    exception.InvalidTemplateReference):
                # This output ain't gonna work anyway
                self._deps = set()
        return self._deps

    def dep_attrs(self, resource_name, load_all=False):
        """Iterate over attributes of a given resource that this references.

        Return an iterator over dependent attributes for specified
        resource_name in the output's value field.
        """
        if self._all_dep_attrs is None and load_all:
            attr_map = collections.defaultdict(set)
            for r, a in function.all_dep_attrs(self._value):
                attr_map[r].add(a)
            self._all_dep_attrs = attr_map

        if self._all_dep_attrs is not None:
            return iter(self._all_dep_attrs.get(resource_name, []))

        return function.dep_attrs(self._value, resource_name)

    def get_value(self):
        """Resolve the value of the output."""
        if self._resolved_value is None:
            self._resolved_value = function.resolve(self._value)
        return self._resolved_value

    def description(self):
        """Return a description of the output."""
        if self._description is None:
            return 'No description given'

        return str(self._description)

    def render_hot(self):
        def items():
            if self._description is not None:
                yield 'description', self._description
            yield 'value', copy.deepcopy(self._value)
        return dict(items())



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import itertools
import weakref

from heat.common import exception
from heat.common.i18n import _


class Function(metaclass=abc.ABCMeta):
    """Abstract base class for template functions."""

    def __init__(self, stack, fn_name, args):
        """Initialise with a Stack, the function name and the arguments.

        All functions take the form of a single-item map in JSON::

            { <fn_name> : <args> }
        """
        super(Function, self).__init__()
        self._stackref = weakref.ref(stack) if stack is not None else None
        self.fn_name = fn_name
        self.args = args

    @property
    def stack(self):
        ref = self._stackref
        if ref is None:
            return None

        stack = ref()
        assert stack is not None, ("Need a reference to the "
                                   "StackDefinition object")
        return stack

    def validate(self):
        """Validate arguments without resolving the function.

        Function subclasses must override this method to validate their
        args.
        """
        validate(self.args)

    @abc.abstractmethod
    def result(self):
        """Return the result of resolving the function.

        Function subclasses must override this method to calculate their
        results.
        """
        return {self.fn_name: self.args}

    def dependencies(self, path):
        return dependencies(self.args, '.'.join([path, self.fn_name]))

    def dep_attrs(self, resource_name):
        """Return the attributes of the specified resource that are referenced.

        Return an iterator over any attributes of the specified resource that
        this function references.

        The special value heat.engine.attributes.ALL_ATTRIBUTES may be used to
        indicate that all attributes of the resource are required.
        """
        return dep_attrs(self.args, resource_name)

    def all_dep_attrs(self):
        """Return resource, attribute name pairs of all attributes referenced.

        Return an iterator over the resource name, attribute name tuples of
        all attributes that this function references.

        The special value heat.engine.attributes.ALL_ATTRIBUTES may be used to
        indicate that all attributes of the resource are required.

        By default this calls the dep_attrs() method, but subclasses can
        override to provide a more efficient implementation.
        """
        # If we are using the default dep_attrs method then it will only
        # return data from the args anyway
        if type(self).dep_attrs == Function.dep_attrs:
            return all_dep_attrs(self.args)

        def res_dep_attrs(resource_name):
            return zip(itertools.repeat(resource_name),
                       self.dep_attrs(resource_name))

        resource_names = self.stack.enabled_rsrc_names()

        return itertools.chain.from_iterable(map(res_dep_attrs,
                                                 resource_names))

    def __reduce__(self):
        """Return a representation of the function suitable for pickling.

        This allows the copy module (which works by pickling and then
        unpickling objects) to copy a template. Functions in the copy will
        return to their original (JSON) form (i.e. a single-element map).
        """
        return dict, ([(self.fn_name, self.args)],)

    def _repr_result(self):
        try:
            return repr(self.result())
        except (TypeError, ValueError):
            return '???'

    def __repr__(self):
        """Return a string representation of the function.

        The representation includes the function name, arguments and result
        (if available), as well as the name of the function class.
        """
        fntype = type(self)
        classname = '.'.join(filter(None,
                                    (getattr(fntype,
                                             attr,
                                             '') for attr in ('__module__',
                                                              '__name__'))))
        return '<%s {%s: %r} -> %s>' % (classname,
                                        self.fn_name, self.args,
                                        self._repr_result())

    def __eq__(self, other):
        """Compare the result of this function for equality."""
        try:
            result = self.result()

            if isinstance(other, Function):
                return result == other.result()
            else:
                return result == other

        except (TypeError, ValueError):
            return NotImplemented

    def __ne__(self, other):
        """Compare the result of this function for inequality."""
        eq = self.__eq__(other)
        if eq is NotImplemented:
            return NotImplemented
        return not eq

    __hash__ = None


class Macro(Function, metaclass=abc.ABCMeta):
    """Abstract base class for template macros.

    A macro differs from a function in that it controls how the template is
    parsed. As such, it operates on the syntax tree itself, not on the parsed
    output.
    """
    def __init__(self, stack, fn_name, raw_args, parse_func, template):
        """Initialise with the argument syntax tree and parser function."""
        super(Macro, self).__init__(stack, fn_name, raw_args)
        self._tmplref = weakref.ref(template) if template is not None else None
        self.parsed = self.parse_args(parse_func)

    @property
    def template(self):
        ref = self._tmplref
        if ref is None:
            return None

        tmpl = ref()
        assert tmpl is not None, "Need a reference to the Template object"
        return tmpl

    @abc.abstractmethod
    def parse_args(self, parse_func):
        """Parse the macro using the supplied parsing function.

        Macro subclasses should override this method to control parsing of
        the arguments.
        """
        return parse_func(self.args)

    def validate(self):
        """Validate arguments without resolving the result."""
        validate(self.parsed)

    def result(self):
        """Return the resolved result of the macro contents."""
        return resolve(self.parsed, nullable=True)

    def dependencies(self, path):
        return dependencies(self.parsed, '.'.join([path, self.fn_name]))

    def dep_attrs(self, resource_name):
        """Return the attributes of the specified resource that are referenced.

        Return an iterator over any attributes of the specified resource that
        this function references.

        The special value heat.engine.attributes.ALL_ATTRIBUTES may be used to
        indicate that all attributes of the resource are required.
        """
        return dep_attrs(self.parsed, resource_name)

    def all_dep_attrs(self):
        """Return resource, attribute name pairs of all attributes referenced.

        Return an iterator over the resource name, attribute name tuples of
        all attributes that this function references.

        The special value heat.engine.attributes.ALL_ATTRIBUTES may be used to
        indicate that all attributes of the resource are required.

        By default this calls the dep_attrs() method, but subclasses can
        override to provide a more efficient implementation.
        """
        # If we are using the default dep_attrs method then it will only
        # return data from the transformed parsed args anyway
        if type(self).dep_attrs == Macro.dep_attrs:
            return all_dep_attrs(self.parsed)

        return super(Macro, self).all_dep_attrs()

    def __reduce__(self):
        """Return a representation of the macro result suitable for pickling.

        This allows the copy module (which works by pickling and then
        unpickling objects) to copy a template. Functions in the copy will
        return to their original (JSON) form (i.e. a single-element map).

        Unlike other functions, macros are *not* preserved during a copy. The
        the processed (but unparsed) output is returned in their place.
        """
        if isinstance(self.parsed, Function):
            return self.parsed.__reduce__()
        if self.parsed is None:
            return lambda x: None, (None,)
        return type(self.parsed), (self.parsed,)

    def _repr_result(self):
        return repr(self.parsed)


def _non_null_item(i):
    k, v = i
    return v is not Ellipsis


def _non_null_value(v):
    return v is not Ellipsis


def resolve(snippet, nullable=False):
    if isinstance(snippet, Function):
        result = snippet.result()
        if not (nullable or _non_null_value(result)):
            result = None
        return result

    if isinstance(snippet, collections.abc.Mapping):
        return dict(filter(_non_null_item,
                           ((k, resolve(v, nullable=True))
                            for k, v in snippet.items())))
    elif (not isinstance(snippet, str) and
          isinstance(snippet, collections.abc.Iterable)):
        return list(filter(_non_null_value,
                           (resolve(v, nullable=True) for v in snippet)))

    return snippet


def validate(snippet, path=None):
    if path is None:
        path = []
    elif isinstance(path, str):
        path = [path]

    if isinstance(snippet, Function):
        try:
            snippet.validate()
        except AssertionError:
            raise
        except Exception as e:
            raise exception.StackValidationFailed(
                path=path + [snippet.fn_name],
                message=str(e))
    elif isinstance(snippet, collections.abc.Mapping):
        for k, v in snippet.items():
            validate(v, path + [k])
    elif (not isinstance(snippet, str) and
          isinstance(snippet, collections.abc.Iterable)):
        basepath = list(path)
        parent = basepath.pop() if basepath else ''
        for i, v in enumerate(snippet):
            validate(v, basepath + ['%s[%d]' % (parent, i)])


def dependencies(snippet, path=''):
    """Return an iterator over Resource dependencies in a template snippet.

    The snippet should be already parsed to insert Function objects where
    appropriate.
    """

    if isinstance(snippet, Function):
        return snippet.dependencies(path)

    elif isinstance(snippet, collections.abc.Mapping):
        def mkpath(key):
            return '.'.join([path, str(key)])

        deps = (dependencies(value,
                             mkpath(key)) for key, value in snippet.items())
        return itertools.chain.from_iterable(deps)

    elif (not isinstance(snippet, str) and
          isinstance(snippet, collections.abc.Iterable)):
        def mkpath(idx):
            return ''.join([path, '[%d]' % idx])

        deps = (dependencies(value,
                             mkpath(i)) for i, value in enumerate(snippet))
        return itertools.chain.from_iterable(deps)

    else:
        return []


def dep_attrs(snippet, resource_name):
    """Iterator over dependent attrs of a resource in a template snippet.

    The snippet should be already parsed to insert Function objects where
    appropriate.

    :returns: an iterator over the attributes of the specified resource that
              are referenced in the template snippet.
    """

    if isinstance(snippet, Function):
        return snippet.dep_attrs(resource_name)

    elif isinstance(snippet, collections.abc.Mapping):
        attrs = (dep_attrs(val, resource_name) for val in snippet.values())
        return itertools.chain.from_iterable(attrs)
    elif (not isinstance(snippet, str) and
          isinstance(snippet, collections.abc.Iterable)):
        attrs = (dep_attrs(value, resource_name) for value in snippet)
        return itertools.chain.from_iterable(attrs)
    return []


def all_dep_attrs(snippet):
    """Iterator over resource, attribute name pairs referenced in a snippet.

    The snippet should be already parsed to insert Function objects where
    appropriate.

    :returns: an iterator over the resource name, attribute name tuples of all
              attributes that are referenced in the template snippet.
    """

    if isinstance(snippet, Function):
        return snippet.all_dep_attrs()

    elif isinstance(snippet, collections.abc.Mapping):
        res_attrs = (all_dep_attrs(value) for value in snippet.values())
        return itertools.chain.from_iterable(res_attrs)
    elif (not isinstance(snippet, str) and
          isinstance(snippet, collections.abc.Iterable)):
        res_attrs = (all_dep_attrs(value) for value in snippet)
        return itertools.chain.from_iterable(res_attrs)
    return []


class Invalid(Function):
    """A function for checking condition functions and to force failures.

    This function is used to force failures for functions that are not
    supported in condition definition.
    """

    def __init__(self, stack, fn_name, args):
        raise ValueError(_('The function "%s" '
                           'is invalid in this context') % fn_name)

    def result(self):
        return super(Invalid, self).result()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import client as cc
from cinderclient import exceptions
from keystoneauth1 import exceptions as ks_exceptions
from oslo_config import cfg
from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine.clients import client_plugin
from heat.engine.clients import os as os_client
from heat.engine import constraints


LOG = logging.getLogger(__name__)

CLIENT_NAME = 'cinder'


class CinderClientPlugin(os_client.ExtensionMixin,
                         client_plugin.ClientPlugin):

    exceptions_module = exceptions

    service_types = [VOLUME_V2, VOLUME_V3] = ['volumev2', 'volumev3']

    def get_volume_api_version(self):
        '''Returns the most recent API version.'''
        self.interface = self._get_client_option(CLIENT_NAME, 'endpoint_type')
        try:
            self.context.keystone_session.get_endpoint(
                service_type=self.VOLUME_V3,
                interface=self.interface)
            self.service_type = self.VOLUME_V3
            self.client_version = '3'
        except ks_exceptions.EndpointNotFound:
            try:
                self.context.keystone_session.get_endpoint(
                    service_type=self.VOLUME_V2,
                    interface=self.interface)
                self.service_type = self.VOLUME_V2
                self.client_version = '2'
            except ks_exceptions.EndpointNotFound:
                raise exception.Error(_('No volume service available.'))

    def _create(self):
        self.get_volume_api_version()
        extensions = cc.discover_extensions(self.client_version)
        args = {
            'session': self.context.keystone_session,
            'extensions': extensions,
            'interface': self.interface,
            'service_type': self.service_type,
            'region_name': self._get_region_name(),
            'connect_retries': cfg.CONF.client_retry_limit,
            'http_log_debug': self._get_client_option(CLIENT_NAME,
                                                      'http_log_debug')
        }

        client = cc.Client(self.client_version, **args)
        return client

    @os_client.MEMOIZE_EXTENSIONS
    def _list_extensions(self):
        extensions = self.client().list_extensions.show_all()
        return set(extension.alias for extension in extensions)

    def get_volume(self, volume):
        try:
            return self.client().volumes.get(volume)
        except exceptions.NotFound:
            raise exception.EntityNotFound(entity='Volume', name=volume)

    def get_volume_snapshot(self, snapshot):
        try:
            return self.client().volume_snapshots.get(snapshot)
        except exceptions.NotFound:
            raise exception.EntityNotFound(entity='VolumeSnapshot',
                                           name=snapshot)

    def get_volume_backup(self, backup):
        try:
            return self.client().backups.get(backup)
        except exceptions.NotFound:
            raise exception.EntityNotFound(entity='Volume backup',
                                           name=backup)

    def get_volume_type(self, volume_type):
        vt_id = None
        volume_type_list = self.client().volume_types.list()
        for vt in volume_type_list:
            if volume_type in [vt.name, vt.id]:
                vt_id = vt.id
                break
        if vt_id is None:
            raise exception.EntityNotFound(entity='VolumeType',
                                           name=volume_type)

        return vt_id

    def get_qos_specs(self, qos_specs):
        try:
            qos = self.client().qos_specs.get(qos_specs)
        except exceptions.NotFound:
            qos = self.client().qos_specs.find(name=qos_specs)
        return qos.id

    def is_not_found(self, ex):
        return isinstance(ex, exceptions.NotFound)

    def is_over_limit(self, ex):
        return isinstance(ex, exceptions.OverLimit)

    def is_conflict(self, ex):
        return (isinstance(ex, exceptions.ClientException) and
                ex.code == 409)

    def check_detach_volume_complete(self, vol_id, server_id=None):
        try:
            vol = self.client().volumes.get(vol_id)
        except Exception as ex:
            self.ignore_not_found(ex)
            return True

        server_ids = [
            a['server_id'] for a in vol.attachments if 'server_id' in a]
        if server_id and server_id not in server_ids:
            return True

        if vol.status in ('in-use', 'detaching', 'reserved'):
            LOG.debug('%s - volume still in use', vol_id)
            return False

        LOG.debug('Volume %(id)s - status: %(status)s', {
            'id': vol.id, 'status': vol.status})

        if vol.status not in ('available', 'deleting'):
            LOG.debug("Detachment failed - volume %(vol)s "
                      "is in %(status)s status",
                      {"vol": vol.id, "status": vol.status})
            raise exception.ResourceUnknownStatus(
                resource_status=vol.status,
                result=_('Volume detachment failed'))
        else:
            return True

    def check_attach_volume_complete(self, vol_id):
        vol = self.client().volumes.get(vol_id)
        if vol.status in ('available', 'attaching', 'reserved'):
            LOG.debug("Volume %(id)s is being attached - "
                      "volume status: %(status)s",
                      {'id': vol_id, 'status': vol.status})
            return False

        if vol.status != 'in-use':
            LOG.debug("Attachment failed - volume %(vol)s is "
                      "in %(status)s status",
                      {"vol": vol_id, "status": vol.status})
            raise exception.ResourceUnknownStatus(
                resource_status=vol.status,
                result=_('Volume attachment failed'))

        LOG.info('Attaching volume %(id)s complete', {'id': vol_id})
        return True


class BaseCinderConstraint(constraints.BaseCustomConstraint):

    resource_client_name = CLIENT_NAME


class VolumeConstraint(BaseCinderConstraint):

    resource_getter_name = 'get_volume'


class VolumeSnapshotConstraint(BaseCinderConstraint):

    resource_getter_name = 'get_volume_snapshot'


class VolumeTypeConstraint(BaseCinderConstraint):

    resource_getter_name = 'get_volume_type'


class VolumeBackupConstraint(BaseCinderConstraint):

    resource_getter_name = 'get_volume_backup'


class QoSSpecsConstraint(BaseCinderConstraint):

    expected_exceptions = (exceptions.NotFound,)

    resource_getter_name = 'get_qos_specs'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_serialization import jsonutils

from heat.common import identifier
from heat.common import template_format
from heat.engine.clients.os import glance
from heat.engine.clients.os import heat_plugin
from heat.engine.clients.os import nova
from heat.engine import environment
from heat.engine.resources.aws.cfn.wait_condition_handle import (
    WaitConditionHandle)
from heat.engine.resources.aws.ec2 import instance
from heat.engine.resources.openstack.nova import server
from heat.engine import scheduler
from heat.engine import service
from heat.engine import stack as stk
from heat.engine import stk_defn
from heat.engine import template as tmpl
from heat.tests import common
from heat.tests import utils


TEST_TEMPLATE_METADATA = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "",
  "Parameters" : {
    "KeyName" : {"Type" : "String", "Default": "mine" },
  },
  "Resources" : {
    "S1": {
      "Type": "AWS::EC2::Instance",
      "Metadata" : {
        "AWS::CloudFormation::Init" : {
          "config" : {
            "files" : {
              "/tmp/random_file" : {
                "content" : { "Fn::Join" : ["", [
                  "s2-ip=", {"Fn::GetAtt": ["S2", "PublicIp"]}
                ]]},
                "mode"    : "000400",
                "owner"   : "root",
                "group"   : "root"
              }
            }
          }
        }
      },
      "Properties": {
        "ImageId"      : "a",
        "InstanceType" : "m1.large",
        "KeyName"      : { "Ref" : "KeyName" },
        "UserData"     : "#!/bin/bash -v\n"
      }
    },
    "S2": {
      "Type": "AWS::EC2::Instance",
      "Properties": {
        "ImageId"      : "a",
        "InstanceType" : "m1.large",
        "KeyName"      : { "Ref" : "KeyName" },
        "UserData"     : "#!/bin/bash -v\n"
      }
    }
  }
}
'''

TEST_TEMPLATE_WAIT_CONDITION = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "Just a WaitCondition.",
  "Parameters" : {
    "KeyName" : {"Type" : "String", "Default": "mine" },
  },
  "Resources" : {
    "WH" : {
      "Type" : "AWS::CloudFormation::WaitConditionHandle"
    },
    "S1": {
      "Type": "AWS::EC2::Instance",
      "Properties": {
        "ImageId"      : "a",
        "InstanceType" : "m1.large",
        "KeyName"      : { "Ref" : "KeyName" },
        "UserData"     : { "Fn::Join" : [ "", [ "#!/bin/bash -v\n",
                                                "echo ",
                                                { "Ref" : "WH" },
                                                "\n" ] ] }
      }
    },
    "WC" : {
      "Type" : "AWS::CloudFormation::WaitCondition",
      "DependsOn": "S1",
      "Properties" : {
        "Handle" : {"Ref" : "WH"},
        "Timeout" : "5"
      }
    },
    "S2": {
      "Type": "AWS::EC2::Instance",
      "Metadata" : {
        "test" : {"Fn::GetAtt": ["WC", "Data"]}
      },
      "Properties": {
        "ImageId"      : "a",
        "InstanceType" : "m1.large",
        "KeyName"      : { "Ref" : "KeyName" },
        "UserData"     : "#!/bin/bash -v\n"
      }
    }
  }
}
'''


TEST_TEMPLATE_SERVER = '''
heat_template_version: 2013-05-23
resources:
  instance1:
    type: OS::Nova::Server
    metadata: {"template_data": {get_attr: [instance2, networks]}}
    properties:
      image: cirros-0.3.2-x86_64-disk
      flavor: m1.small
      key_name: stack_key
  instance2:
    type: OS::Nova::Server
    metadata: {'apples': 'pears'}
    properties:
      image: cirros-0.3.2-x86_64-disk
      flavor: m1.small
      key_name: stack_key
'''


class MetadataRefreshTest(common.HeatTestCase):

    @mock.patch.object(nova.NovaClientPlugin, 'find_flavor_by_name_or_id')
    @mock.patch.object(glance.GlanceClientPlugin, 'find_image_by_name_or_id')
    @mock.patch.object(instance.Instance, 'handle_create')
    @mock.patch.object(instance.Instance, 'check_create_complete')
    @mock.patch.object(instance.Instance, '_resolve_attribute')
    def test_FnGetAtt_metadata_updated(self, mock_get, mock_check,
                                       mock_handle, *args):
        """Tests that metadata gets updated when FnGetAtt return changes."""
        # Setup
        temp = template_format.parse(TEST_TEMPLATE_METADATA)
        template = tmpl.Template(temp,
                                 env=environment.Environment({}))
        ctx = utils.dummy_context()
        stack = stk.Stack(ctx, 'test_stack', template, disable_rollback=True)
        stack.store()

        self.stub_KeypairConstraint_validate()

        # Configure FnGetAtt to return different values on subsequent calls
        mock_get.side_effect = [
            '10.0.0.1',
            '10.0.0.2',
        ]

        # Initial resolution of the metadata
        stack.create()
        self.assertEqual((stack.CREATE, stack.COMPLETE), stack.state)

        # Sanity check on S2
        s2 = stack['S2']
        self.assertEqual((s2.CREATE, s2.COMPLETE), s2.state)

        # Verify S1 is using the initial value from S2
        s1 = stack['S1']
        content = self._get_metadata_content(s1.metadata_get())
        self.assertEqual('s2-ip=10.0.0.1', content)

        # This is not a terribly realistic test - the metadata updates below
        # happen in run_alarm_action() in service_stack_watch, and actually
        # operate on a freshly loaded stack so there's no cached attributes.
        # Clear the attributes cache here to keep it passing.
        s2.attributes.reset_resolved_values()

        # Run metadata update to pick up the new value from S2
        # (simulating run_alarm_action() in service_stack_watch)
        s2.metadata_update()
        stk_defn.update_resource_data(stack.defn, s2.name, s2.node_data())
        s1.metadata_update()
        stk_defn.update_resource_data(stack.defn, s1.name, s1.node_data())

        # Verify the updated value is correct in S1
        content = self._get_metadata_content(s1.metadata_get())
        self.assertEqual('s2-ip=10.0.0.2', content)

        # Verify outgoing calls
        mock_get.assert_has_calls([
            mock.call('PublicIp'),
            mock.call('PublicIp')])
        self.assertEqual(2, mock_handle.call_count)
        self.assertEqual(2, mock_check.call_count)

    @staticmethod
    def _get_metadata_content(m):
        tmp = m['AWS::CloudFormation::Init']['config']['files']
        return tmp['/tmp/random_file']['content']


class WaitConditionMetadataUpdateTest(common.HeatTestCase):

    def setUp(self):
        super(WaitConditionMetadataUpdateTest, self).setUp()
        self.man = service.EngineService('a-host', 'a-topic')

    @mock.patch.object(nova.NovaClientPlugin, 'find_flavor_by_name_or_id')
    @mock.patch.object(glance.GlanceClientPlugin, 'find_image_by_name_or_id')
    @mock.patch.object(heat_plugin.HeatClientPlugin, 'get_heat_cfn_url')
    @mock.patch.object(instance.Instance, 'handle_create')
    @mock.patch.object(instance.Instance, 'check_create_complete')
    @mock.patch.object(scheduler.TaskRunner, '_sleep')
    @mock.patch.object(WaitConditionHandle, 'identifier')
    def test_wait_metadata(self, mock_identifier, mock_sleep,
                           mock_check, mock_handle, mock_get, *args):
        """Tests a wait condition metadata update after a signal call."""

        mock_get.return_value = 'http://server.test:8000/v1'
        # Setup Stack
        temp = template_format.parse(TEST_TEMPLATE_WAIT_CONDITION)
        template = tmpl.Template(temp)
        ctx = utils.dummy_context()
        stack = stk.Stack(ctx, 'test-stack', template, disable_rollback=True)
        stack.store()

        self.stub_KeypairConstraint_validate()

        res_id = identifier.ResourceIdentifier('test_tenant_id', stack.name,
                                               stack.id, '', 'WH')
        mock_identifier.return_value = res_id

        watch = stack['WC']
        inst = stack['S2']

        # Setup Sleep Behavior
        self.run_empty = True

        def check_empty(sleep_time):
            self.assertEqual('{}', watch.FnGetAtt('Data'))
            self.assertIsNone(inst.metadata_get()['test'])

        def update_metadata(unique_id, data, reason):
            self.man.resource_signal(ctx,
                                     dict(stack.identifier()),
                                     'WH',
                                     {'Data': data, 'Reason': reason,
                                      'Status': 'SUCCESS',
                                      'UniqueId': unique_id},
                                     sync_call=True)

        def post_success(sleep_time):
            update_metadata('123', 'foo', 'bar')

        def side_effect_popper(sleep_time):
            wh = stack['WH']
            if wh.status == wh.IN_PROGRESS:
                return
            elif self.run_empty:
                self.run_empty = False
                check_empty(sleep_time)
            else:
                post_success(sleep_time)

        mock_sleep.side_effect = side_effect_popper

        # Test Initial Creation
        stack.create()
        self.assertEqual((stack.CREATE, stack.COMPLETE), stack.state)
        self.assertEqual('{"123": "foo"}', watch.FnGetAtt('Data'))
        self.assertEqual('{"123": "foo"}', inst.metadata_get()['test'])

        # Test Update
        update_metadata('456', 'blarg', 'wibble')

        self.assertEqual({'123': 'foo', '456': 'blarg'},
                         jsonutils.loads(watch.FnGetAtt('Data')))
        self.assertEqual('{"123": "foo"}',
                         inst.metadata_get()['test'])
        self.assertEqual(
            {'123': 'foo', '456': 'blarg'},
            jsonutils.loads(inst.metadata_get(refresh=True)['test']))

        # Verify outgoing calls
        self.assertEqual(2, mock_handle.call_count)
        self.assertEqual(2, mock_check.call_count)


class MetadataRefreshServerTest(common.HeatTestCase):

    @mock.patch.object(nova.NovaClientPlugin, 'find_flavor_by_name_or_id',
                       return_value=1)
    @mock.patch.object(glance.GlanceClientPlugin, 'find_image_by_name_or_id',
                       return_value=1)
    @mock.patch.object(server.Server, 'handle_create')
    @mock.patch.object(server.Server, 'check_create_complete')
    @mock.patch.object(server.Server, 'get_attribute', new_callable=mock.Mock)
    def test_FnGetAtt_metadata_update(self, mock_get, mock_check,
                                      mock_handle, *args):
        temp = template_format.parse(TEST_TEMPLATE_SERVER)
        template = tmpl.Template(temp,
                                 env=environment.Environment({}))
        ctx = utils.dummy_context()
        stack = stk.Stack(ctx, 'test-stack', template, disable_rollback=True)
        stack.store()

        self.stub_KeypairConstraint_validate()

        # Note dummy addresses are from TEST-NET-1 ref rfc5737
        mock_get.side_effect = ['192.0.2.1', '192.0.2.2']

        # Test
        stack.create()
        self.assertEqual((stack.CREATE, stack.COMPLETE), stack.state)

        s1 = stack['instance1']
        s2 = stack['instance2']
        md = s1.metadata_get()
        self.assertEqual({u'template_data': '192.0.2.1'}, md)

        # Now set some metadata via the resource, like is done by
        # _populate_deployments_metadata. This should be persisted over
        # calls to metadata_update()
        new_md = {u'template_data': '192.0.2.2', 'set_by_rsrc': 'orange'}
        s1.metadata_set(new_md)
        md = s1.metadata_get(refresh=True)
        self.assertEqual(new_md, md)
        s2.attributes.reset_resolved_values()
        stk_defn.update_resource_data(stack.defn, s2.name, s2.node_data())
        s1.metadata_update()
        md = s1.metadata_get(refresh=True)
        self.assertEqual(new_md, md)

        # Verify outgoing calls
        mock_get.assert_has_calls([
            mock.call('networks'),
            mock.call('networks')])
        self.assertEqual(2, mock_handle.call_count)
        self.assertEqual(2, mock_check.call_count)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutronclient.neutron import v2_0 as neutronV20
from osc_lib import exceptions

from heat.common import template_format
from heat.engine.resources.openstack.octavia import pool_member
from heat.tests import common
from heat.tests.openstack.octavia import inline_templates
from heat.tests import utils


class PoolMemberTest(common.HeatTestCase):

    def test_resource_mapping(self):
        mapping = pool_member.resource_mapping()
        self.assertEqual(pool_member.PoolMember,
                         mapping['OS::Octavia::PoolMember'])

    def _create_stack(self, tmpl=inline_templates.MEMBER_TEMPLATE):
        self.t = template_format.parse(tmpl)
        self.stack = utils.parse_stack(self.t)
        self.member = self.stack['member']
        self.patchobject(neutronV20, 'find_resourceid_by_name_or_id',
                         return_value='123')

        self.octavia_client = mock.MagicMock()
        self.member.client = mock.MagicMock(return_value=self.octavia_client)
        self.member.client_plugin().get_pool = (
            mock.MagicMock(return_value='123'))
        self.member.client_plugin().client = mock.MagicMock(
            return_value=self.octavia_client)
        self.member.translate_properties(self.member.properties)

    def test_create(self):
        self._create_stack()
        self.octavia_client.member_show.side_effect = [
            {'provisioning_status': 'PENDING_CREATE'},
            {'provisioning_status': 'PENDING_CREATE'},
            {'provisioning_status': 'ACTIVE'},
        ]
        self.octavia_client.member_create.side_effect = [
            exceptions.Conflict(409), {'member': {'id': '1234'}}]
        expected = {
            'member': {
                'address': '1.2.3.4',
                'protocol_port': 80,
                'weight': 1,
                'subnet_id': '123',
                'admin_state_up': True,
                'tags': ['tag1', 'tag2']
            }
        }
        props = self.member.handle_create()
        self.assertFalse(self.member.check_create_complete(props))
        self.octavia_client.member_create.assert_called_with('123',
                                                             json=expected)
        self.assertFalse(self.member.check_create_complete(props))
        self.octavia_client.member_create.assert_called_with('123',
                                                             json=expected)
        self.assertFalse(self.member.check_create_complete(props))
        self.assertTrue(self.member.check_create_complete(props))

    def test_show_resource(self):
        self._create_stack()
        self.member.resource_id_set('1234')
        self.octavia_client.member_show.return_value = {'id': '1234'}

        self.assertEqual(self.member._show_resource(), {'id': '1234'})

        self.octavia_client.member_show.assert_called_with('123', '1234')

    def test_update(self):
        self._create_stack()
        self.member.resource_id_set('1234')
        self.octavia_client.member_show.side_effect = [
            {'provisioning_status': 'PENDING_UPDATE'},
            {'provisioning_status': 'PENDING_UPDATE'},
            {'provisioning_status': 'ACTIVE'},
        ]
        self.octavia_client.member_set.side_effect = [
            exceptions.Conflict(409), None]
        prop_diff = {
            'admin_state_up': False,
            'weight': 2,
            'tags': ['tag3', 'tag4']
        }

        prop_diff = self.member.handle_update(None, None, prop_diff)

        self.assertFalse(self.member.check_update_complete(prop_diff))
        self.assertFalse(self.member._update_called)
        self.octavia_client.member_set.assert_called_with(
            '123', '1234', json={'member': prop_diff})
        self.assertFalse(self.member.check_update_complete(prop_diff))
        self.assertTrue(self.member._update_called)
        self.octavia_client.member_set.assert_called_with(
            '123', '1234', json={'member': prop_diff})
        self.assertFalse(self.member.check_update_complete(prop_diff))
        self.assertTrue(self.member.check_update_complete(prop_diff))

    def test_delete(self):
        self._create_stack()
        self.member.resource_id_set('1234')
        self.octavia_client.member_show.side_effect = [
            {'provisioning_status': 'PENDING_DELETE'},
            {'provisioning_status': 'PENDING_DELETE'},
            {'provisioning_status': 'DELETED'},
        ]
        self.octavia_client.member_delete.side_effect = [
            exceptions.Conflict(409),
            None]

        self.member.handle_delete()

        self.assertFalse(self.member.check_delete_complete(None))
        self.assertFalse(self.member._delete_called)
        self.octavia_client.member_delete.assert_called_with('123',
                                                             '1234')
        self.assertFalse(self.member.check_delete_complete(None))
        self.octavia_client.member_delete.assert_called_with('123',
                                                             '1234')
        self.assertTrue(self.member._delete_called)
        self.assertTrue(self.member.check_delete_complete(None))

    def test_delete_not_found(self):
        self._create_stack()
        self.member.resource_id_set('1234')
        self.octavia_client.member_show.side_effect = [
            {'provisioning_status': 'PENDING_DELETE'},
        ]
        self.octavia_client.member_delete.side_effect = [
            exceptions.Conflict(409),
            exceptions.NotFound(404)]

        self.member.handle_delete()

        self.assertFalse(self.member.check_delete_complete(None))
        self.assertFalse(self.member._delete_called)
        self.octavia_client.member_delete.assert_called_with('123',
                                                             '1234')
        self.assertTrue(self.member.check_delete_complete(None))
        self.octavia_client.member_delete.assert_called_with('123',
                                                             '1234')
        self.assertFalse(self.member._delete_called)

    def test_delete_with_pool_not_found(self):
        self._create_stack()
        self.member.resource_id_set('1234')
        m_get_pool = mock.Mock(side_effect=exceptions.NotFound(404))
        self.member.client_plugin().get_pool = m_get_pool
        self.octavia_client.member_delete.side_effect = [
            exceptions.NotFound(404)]
        self.member.translate_properties(self.member.properties,
                                         ignore_resolve_error=True)
        self.member.handle_delete()
        self.assertTrue(self.member.check_delete_complete(None))
        self.octavia_client.member_delete.assert_called_with('123',
                                                             '1234')

    def test_delete_failed(self):
        self._create_stack()
        self.member.resource_id_set('1234')
        self.octavia_client.member_delete.side_effect = (
            exceptions.Unauthorized(401))

        self.member.handle_delete()

        self.assertRaises(exceptions.Unauthorized,
                          self.member.check_delete_complete, None)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from octaviaclient.api import constants
from octaviaclient.api.v2 import octavia
from osc_lib import exceptions

from heat.engine.clients import client_plugin
from heat.engine import constraints

CLIENT_NAME = 'octavia'
DEFAULT_FIND_ATTR = 'name'


def _is_translated_exception(ex, code):
    return (isinstance(ex, octavia.OctaviaClientException)
            and ex.code == code)


class OctaviaClientPlugin(client_plugin.ClientPlugin):

    exceptions_module = octavia

    service_types = [LOADBALANCER] = ['load-balancer']

    supported_versions = [V2] = ['2']

    default_version = V2

    def _create(self, version=None):
        interface = self._get_client_option(CLIENT_NAME, 'endpoint_type')
        endpoint = self.url_for(service_type=self.LOADBALANCER,
                                endpoint_type=interface)
        return octavia.OctaviaAPI(
            session=self.context.keystone_session,
            service_type=self.LOADBALANCER,
            endpoint=endpoint)

    def is_not_found(self, ex):
        return isinstance(
            ex, exceptions.NotFound) or _is_translated_exception(ex, 404)

    def is_over_limit(self, ex):
        return isinstance(
            ex, exceptions.OverLimit) or _is_translated_exception(ex, 413)

    def is_conflict(self, ex):
        return isinstance(
            ex, exceptions.Conflict) or _is_translated_exception(ex, 409)

    def get_pool(self, value):
        pool = self.client().find(path=constants.BASE_POOL_URL,
                                  value=value, attr=DEFAULT_FIND_ATTR)
        return pool['id']

    def get_listener(self, value):
        lsnr = self.client().find(path=constants.BASE_LISTENER_URL,
                                  value=value, attr=DEFAULT_FIND_ATTR)
        return lsnr['id']

    def get_loadbalancer(self, value):
        lb = self.client().find(path=constants.BASE_LOADBALANCER_URL,
                                value=value, attr=DEFAULT_FIND_ATTR)
        return lb['id']

    def get_l7policy(self, value):
        policy = self.client().find(path=constants.BASE_L7POLICY_URL,
                                    value=value, attr=DEFAULT_FIND_ATTR)
        return policy['id']

    def get_flavor(self, value):
        flavor = self.client().find(path=constants.BASE_FLAVOR_URL,
                                    value=value, attr=DEFAULT_FIND_ATTR)
        return flavor['id']

    def get_flavorprofile(self, value):
        flavorprofile = self.client().find(
            path=constants.BASE_FLAVORPROFILE_URL,
            value=value, attr=DEFAULT_FIND_ATTR)
        return flavorprofile['id']


class OctaviaConstraint(constraints.BaseCustomConstraint):

    expected_exceptions = (exceptions.NotFound,
                           octavia.OctaviaClientException)
    base_url = None

    def validate_with_client(self, client, value):
        octavia_client = client.client(CLIENT_NAME)
        octavia_client.find(path=self.base_url, value=value,
                            attr=DEFAULT_FIND_ATTR)


class LoadbalancerConstraint(OctaviaConstraint):
    base_url = constants.BASE_LOADBALANCER_URL


class ListenerConstraint(OctaviaConstraint):
    base_url = constants.BASE_LISTENER_URL


class PoolConstraint(OctaviaConstraint):
    base_url = constants.BASE_POOL_URL


class L7PolicyConstraint(OctaviaConstraint):
    base_url = constants.BASE_L7POLICY_URL


class FlavorConstraint(OctaviaConstraint):
    base_url = constants.BASE_FLAVOR_URL


class FlavorProfileConstraint(OctaviaConstraint):
    base_url = constants.BASE_FLAVORPROFILE_URL
